import inspect
import json
import sys
import requests

from future.backports.urllib.parse import urlparse

from otest.check import Check
from otest.check import ERROR
from otest.check import WARNING
from otest.events import EV_PROTOCOL_REQUEST
from otest.events import NoSuchEvent
from otest.shannon_entropy import calculate

from oic.oauth2 import AuthorizationRequest
from oic.oauth2 import AccessTokenRequest
from oic.extension.client import RegistrationRequest
from oic.oic import message

__author__ = 'roland'


def get_message(conv, classes):
    for msg_cls in classes:
        try:
            request = conv.events.get_message(EV_PROTOCOL_REQUEST, msg_cls)
        except NoSuchEvent:
            pass
        else:
            return request

    raise NoSuchEvent(classes)


def registration_request(conv):
    return get_message(conv, [RegistrationRequest, message.RegistrationRequest])


def authorization_request(conv):
    return get_message(conv, [AuthorizationRequest,
                              message.AuthorizationRequest])


def access_token_request(conv):
    return get_message(conv, [AccessTokenRequest,
                              message.AccessTokenRequest])


class VerifyRegistrationOfflineAccess(Check):
    cid = 'verify-registration-offline-access'
    msg = "Check if offline access is requested"

    def _func(self, conv):
        try:
            request = registration_request(conv)
        except NoSuchEvent:
            self._status = ERROR
        else:
            # 'offline_access only allow if response_type == 'code'
            try:
                req_scope = request['scope']
            except KeyError:
                pass
            else:
                if 'offline_access' in req_scope:
                    if request['response_type'] != 'code':
                        self._status = ERROR
                        self._message = 'Offline access not allowed for ' \
                                        'anything but code flow'

        return {}


class VerifyRegistrationResponseTypes(Check):
    cid = 'verify-registration-response_types'
    msg = "Only one of 'code' or 'token' allowed"

    def _func(self, conv):
        try:
            request = registration_request(conv)
        except NoSuchEvent:
            self._status = ERROR
        else:
            try:
                resp_types = request['response_types']
            except KeyError:
                pass
            else:
                try:
                    _allowed = self._kwargs['allowed']
                except KeyError:
                    if len(resp_types) > 1:
                        self._status = WARNING
                        self._message = 'Only allowed to register one ' \
                                        'response type'
                    elif resp_types[0] not in ['code', 'token']:
                        self._status = ERROR
                        self._message = 'Not allowed response type'
                else:
                    if not set(resp_types).issubset(set(_allowed)):
                        self._status = ERROR
                        self._message = \
                            'Asked for response_types not subset of {}'.format(
                                _allowed
                            )

        return {}


class VerifyRegistrationSoftwareStatement(Check):
    cid = 'verify-registration-software-statement'
    msg = "Verify that the correct claims appear in the Software statement"

    def _func(self, conv):
        try:
            request = registration_request(conv)
        except NoSuchEvent:
            self._status = ERROR
        else:
            try:
                _ss = request['software_statement']
            except KeyError:
                pass
            else:
                missing = []
                for claim in ['redirect_uris', 'grant_types', 'client_name',
                              'client_uri']:
                    if claim not in _ss:
                        missing.append(claim)
                if 'jwks_uri' not in _ss and 'jwks' not in _ss:
                    missing.append('jwks_uri/jwks')

                if missing:
                    self._status = WARNING
                    self._message = 'Missing "{}" claims from Software ' \
                                    'Statement'.format(missing)

        return {}


class VerifyRegistrationRedirectUriScheme(Check):
    cid = 'verify-registration-redirect_uri-scheme'
    msg = "Only certain redirect_uri schemes are allowed"

    def _func(self, conv):
        try:
            request = registration_request(conv)
        except NoSuchEvent:
            self._status = ERROR
        else:
            try:
                ruris = request['redirect_uris']
            except KeyError:
                self._status = ERROR
                self._message = 'MUST register redirect_uris'
            else:
                for ruri in ruris:
                    p = urlparse(ruri)
                    if p.scheme == 'https':
                        continue
                    elif p.scheme == 'http':
                        if 'localhost' != p.netloc.split('.'):
                            self._status = ERROR
                            self._message = 'Not allowed response type'
                            break
                    else:  # How do I check for local schemes ?
                        try:
                            uri_scheme = conv.data['uri_schemes']
                        except KeyError:
                            pass
                        else:
                            for scheme, desc in uri_scheme.items():
                                if p.scheme == scheme:
                                    self._status = WARNING
                                    self._message = \
                                        "None-local URI scheme: {}".format(
                                            scheme)
                                break

        return {}


class VerifyRegistrationPublicKeyRegistration(Check):
    cid = 'verify-registration-public_key-registration'
    msg = "Public key must be registered"

    def _func(self, conv):
        try:
            request = registration_request(conv)
        except NoSuchEvent:
            self._status = ERROR
        else:
            try:
                _uri = request['jwks_uri']
            except KeyError:
                try:
                    jwks = request['jwks']
                except KeyError:
                    self._status = ERROR
                    self._message = 'Must register a public key'
                else:
                    pub = 0
                    for desc in jwks['keys']:
                        if desc['kty'] in ['RSA', 'EC']:
                            pub += 1
                    if pub == 0:
                        self._status = ERROR
                        self._message = 'Must register a public key'
            else:
                resp = requests.request('GET', _uri, verify=False)
                if resp.status_code == 200:
                    jwks = json.loads(resp.text)
                    pub = 0
                    for desc in jwks['keys']:
                        if desc['kty'] in ['RSA', 'EC']:
                            pub += 1
                    if pub == 0:
                        self._status = ERROR
                        self._message = 'Must register a public key'
                else:
                    self._status = ERROR
                    self._message = 'Failed to access the RP keys at {}'.format(
                        _uri)

        return {}


class VerifyAuthorizationOfflineAccess(Check):
    cid = 'verify-authorization-offline-access'
    msg = "Check if offline access is requested"

    def _func(self, conv):
        try:
            request = registration_request(conv)
        except NoSuchEvent:
            self._status = ERROR
        else:
            try:
                req_scopes = request['scope']
            except KeyError:
                pass
            else:
                if 'offline_access' in req_scopes:
                    if request['response_type'] != ['code']:
                        self._status = ERROR
                        self._message = 'Offline access only when using ' \
                                        '"code" flow'

        return {}


class VerifyAuthorizationStateEntropy(Check):
    cid = 'verify-authorization-state-entropy'
    msg = "Check if offline access is requested"

    def _func(self, conv):
        try:
            request = authorization_request(conv)
        except NoSuchEvent:
            self._status = ERROR
            self._message = "No AuthorizationRequest"
        else:
            bits = calculate(request['state'])
            if bits < 128:
                self._status = WARNING
                self._message = 'Not enough entropy in string: {} < 128'.format(
                    bits)
        return {}


class VerifyAuthorizationRedirectUri(Check):
    cid = 'verify-authorization-redirect_uri'
    msg = "Check if offline access is requested"

    def _func(self, conv):
        try:
            clireq_request = registration_request(conv)
        except NoSuchEvent:
            self._status = ERROR
        else:
            try:
                authz_request = authorization_request(conv)
            except NoSuchEvent:
                self._status = ERROR
                self._message = 'No AuthorizationRequest'
            else:
                if authz_request['redirect_uri'] not in clireq_request[
                    'redirect_uris']:
                    self._status = ERROR
                    self._message = 'Redirect_uri not registered'

        return {}


class VerifyTokenRequestClientAssertion(Check):
    cid = 'verify-token-request-client_assertion'
    msg = "Check that the client_assertion JWT contains expected claims"

    def _func(self, conv):
        request = access_token_request(conv)

        ca = request['parsed_client_assertion']
        missing = []
        for claim in ["iss", "sub", "aud", "iat", "exp", "jti"]:
            if claim not in ca:
                missing.append(claim)

        if missing:
            self._status = ERROR
            self._message = 'Redirect_uri not registered'

        # verify jti entropy
        bits = calculate(ca['jti'])
        if bits < 128:
            self._status = WARNING
            self._message = 'Not enough entropy in string: {} < 128'.format(
                bits)

        return {}


def factory(cid):
    for name, obj in inspect.getmembers(sys.modules[__name__]):
        if inspect.isclass(obj) and issubclass(obj, Check):
            try:
                if obj.cid == cid:
                    return obj
            except AttributeError:
                pass

    from otest import check
    return check.factory(cid)



import json
import logging
#from otest.interaction import Interaction
from otest.events import Events

__author__ = 'roland'

logger = logging.getLogger(__name__)


class Conversation(object):
    def __init__(self, flow, entity, msg_factory, check_factory=None,
                 features=None, opid=None, **extra_args):
        self.flow = flow
        self.entity = entity
        self.msg_factory = msg_factory
        self.events = Events()
        #self.interaction = Interaction(self.entity, interaction)
        self.check_factory = check_factory
        self.features = features
        self.operator_id = opid
        self.extra_args = extra_args
        self.test_id = ""
        self.info = {}
        self.index = 0
        self.comhandler = None
        self.exception = None
        self.sequence = []
        self.cache = {}
        self.tool_config = {}
        self.conf = None

    def dump_state(self, filename):
        state = {
            "client": {
                "behaviour": self.entity.behaviour,
                "keyjar": self.entity.keyjar.dump(),
                "provider_info": self.entity.provider_info.to_json(),
                "client_id": self.entity.client_id,
                "client_secret": self.entity.client_secret,
            },
            "sequence": self.flow,
            "flow_index": self.index,
            "client_config": self.entity.conf,
            "condition": self.events.get('condition')
        }

        try:
            state["client"][
                "registration_resp"] = \
                self.entity.registration_response.to_json()
        except AttributeError:
            pass

        txt = json.dumps(state)
        _fh = open(filename, "w")
        _fh.write(txt)
        _fh.close()

    def get_tool_attribute(self, *attr, **kwargs):
        """
        Return the tool configuration attribute value.
        If more then one attribute is specified, first try the first one
        if that doesn'e succeed take try the next one and so on.

        :param attr: A list of attributes
        :param default: If none of the attributes have a value return this
        :return: An attribute value or the default value. If no attribute 
        value or default value raise KeyError.
        """
        for claim in attr:
            try:
                return self.tool_config[claim]
            except KeyError:
                pass

        return kwargs['default']

    def get_redirect_uris(self):
        try:
            return self.entity.registration_response["redirect_uris"]
        except KeyError:
            try:
                return self.entity.registration_info["redirect_uris"]
            except KeyError:
                try:
                    return self.conf.CLIENT['registration_info'][
                        "redirect_uris"]
                except KeyError:
                    return self.conf.CLIENT['registration_response'][
                        "redirect_uris"]



PMAP = {
    "C": "Basic (code)", "I": "Implicit (id_token)",
    "IT": "Implicit (id_token+token)",
    "CI": "Hybrid (code+id_token)", "CT": "Hybrid (code+token)",
    "CIT": "Hybrid (code+id_token+token)"
}



import logging
import os

from future.backports.urllib.parse import parse_qs
from oic.oauth2 import AuthorizationErrorResponse
from oic.oauth2 import AuthorizationRequest
from oic.oauth2 import PyoidcError
from oic.utils.http_util import Redirect
from oic.utils.http_util import Response

from otest import ConditionError
from otest import ConfigurationError
from otest import Done
from otest import exception_trace
from otest import tool
from otest.check import OK
from otest.check import State
from otest.conversation import Conversation
from otest.events import EV_CONDITION
from otest.events import EV_OPERATION
from otest.events import EV_PROTOCOL_REQUEST
from otest.events import EV_REQUEST
from otest.events import EV_RESPONSE
from otest.result import Result
from otest.result import safe_path
from otest.verify import Verify

logger = logging.getLogger(__name__)


class WebTester(tool.Tester):
    def __init__(self, *args, **kwargs):
        tool.Tester.__init__(self, *args, **kwargs)
        try:
            self.base_url = self.conv.entity.base_url
        except AttributeError:
            self.base_url = self.kwargs['base']
        self.provider_cls = self.kwargs['provider_cls']
        self.selected = {}

    def fname(self, test_id):
        _pname = '_'.join(self.profile)
        try:
            return safe_path(self.conv.entity_id, _pname, test_id)
        except (AttributeError, KeyError):
            return safe_path('dummy', _pname, test_id)

    def match_profile(self, test_id, **kwargs):
        _spec = self.flows[test_id]
        # There must be an intersection between the two profile lists.
        if self.sh.profile in _spec["usage"]["return_type"]:
            return True
        else:
            return False

    def setup(self, test_id, **kw_args):
        if not self.match_profile(test_id):
            return False

        self.sh.session_setup(path=test_id)
        _flow = self.flows[test_id]
        try:
            _cap = kw_args['op_profiles'][self.sh['test_conf']['profile']]
        except KeyError:
            _cap = None
        _ent = self.provider_cls(capabilities=_cap, **kw_args['as_args'])
        _ent.baseurl = os.path.join(_ent.baseurl, kw_args['sid'])
        _ent.jwks_uri = os.path.join(_ent.baseurl,
                                     kw_args['as_args']['jwks_name'])
        _ent.name = _ent.baseurl
        self.conv = Conversation(_flow, _ent,
                                 msg_factory=kw_args["msg_factory"])
        self.conv.sequence = self.sh["sequence"]
        _ent.conv = self.conv
        _ent.events = self.conv.events
        self.sh["conv"] = self.conv
        return True

    def run(self, test_id, **kw_args):
        if not self.setup(test_id, **kw_args):
            raise ConfigurationError()

        # noinspection PyTypeChecker
        try:
            return self.run_item(test_id, index=0, **kw_args)
        except Exception as err:
            exception_trace("", err, logger)
            return self.inut.err_response("run", err)

    def post_op(self, oper, res, test_id):
        """
        should be done as late as possible, so all processing has been
        :param oper:
        :return:
        """
        try:
            oper.post_tests()
        except ConditionError:
            pass

        self.conv.events.store(EV_CONDITION, State('Done', OK))
        self.store_result(res)

    def get_cls_and_func(self, index):
        item = self.conv.sequence[index]

        if isinstance(item, tuple):
            cls, funcs = item
        else:
            cls = item
            funcs = {}

        return cls, funcs

    def run_item(self, test_id, index, profiles=None, **kw_args):
        logger.info("<=<=<=<=< %s >=>=>=>=>" % test_id)

        _ss = self.sh
        try:
            _ss.test_flows.complete[test_id] = False
        except KeyError:
            pass

        self.conv.test_id = test_id
        res = Result(self.sh, self.kwargs['profile_handler'])

        if index >= len(self.conv.sequence):
            return None

        try:
            internal = kw_args['internal']
        except KeyError:
            internal = True

        cls, funcs = self.get_cls_and_func(index)

        try:
            _name = cls.__name__
        except AttributeError:
            _name = 'none'
        logger.info("<--<-- {} --- {} -->-->".format(index, _name))
        self.conv.events.store(EV_OPERATION, _name, sender='run_flow')
        try:
            _oper = cls(conv=self.conv, inut=self.inut, sh=self.sh,
                        profile=self.profile, test_id=test_id,
                        funcs=funcs, check_factory=self.chk_factory,
                        cache=self.cache, internal=internal)
            # self.conv.operation = _oper
            if profiles:
                profile_map = profiles.PROFILEMAP
            else:
                profile_map = None
            _oper.setup(profile_map)
            resp = _oper()
        except ConditionError:
            self.store_result(res)
            return False
        except Exception as err:
            exception_trace('run_flow', err)
            self.sh["index"] = index
            return self.inut.err_response("run_sequence", err)
        else:
            if isinstance(resp, self.response_cls):
                if self.conv.sequence[index+1] == Done:
                    self.post_op(_oper, res, test_id)
                return resp

            if resp:
                if self.conv.sequence[index+1] == Done:
                    self.post_op(_oper, res, test_id)
                return resp

        # should be done as late as possible, so all processing has been
        # done
        try:
            _oper.post_tests()
        except ConditionError:
            self.store_result(res)
            return False

        _ss['index'] = self.conv.index = index + 1

        return True

    def display_test_list(self, **kwargs):
        try:
            if self.sh.session_init():
                return self.inut.flow_list()
            else:
                try:
                    resp = Redirect("%s/opresult#%s" % (
                        self.base_url, self.sh["testid"][0]))
                except KeyError:
                    return self.inut.flow_list(**kwargs)
                else:
                    return resp(self.inut.environ, self.inut.start_response)
        except Exception as err:
            exception_trace("display_test_list", err)
            return self.inut.err_response("session_setup", err)

    def handle_request(self, req, path=''):
        logging.debug('Raw request: {}'.format(req))
        if req:
            self.conv.events.store(EV_REQUEST, req)
            func = getattr(self.conv.entity.server,
                           'parse_{}_request'.format(path))

            msg = None
            try:
                if req[0] in ['{', '[']:
                    msg = func(req, sformat='json')
                else:
                    if path in ['authorization', 'check_session']:
                        msg = func(query=req)  # default urlencoded
                    elif path in ['token', 'refresh_token']:
                        msg = func(body=req)
                    else:
                        msg = func(req)
            except PyoidcError as err:
                logging.error('{}'.format(err))

            if msg:
                self.conv.events.store(EV_PROTOCOL_REQUEST, msg)

    def do_config(self, sid='', start_page='', params='', **args):
        resp = Response(mako_template="config.mako",
                        template_lookup=self.kwargs['lookup'], headers=[])

        if sid:
            _url = os.path.join(self.base_url, sid)
        else:
            _url = self.base_url

        try:
            test_id = args['test_id']
        except KeyError:
            test_id = ''

        kwargs = {
            'start_page': start_page,
            'params': params,
            'issuer': _url,
            'profiles': list(self.kwargs['op_profiles'].keys()),
            'selected': self.selected,
            'sid':sid,
            'base': self.base_url,
            'test_id': test_id
        }
        return resp(self.inut.environ, self.inut.start_response, **kwargs)

    def do_next(self, req, filename, path='', **kwargs):
        sh = self.sh

        self.conv = sh['conv']
        cls, funcs = self.get_cls_and_func(self.conv.index+1)
        if cls.endpoint != path:
            if path == 'authorization':  # Jumping the gun here
                areq = AuthorizationRequest().from_urlencoded(req)
                # send an error back to the redirect_uri
                msg = AuthorizationErrorResponse(error='access_denied',
                                                 state=areq['state'])
                _entity = self.conv.entity
                redirect_uri = _entity.get_redirect_uri(areq)
                _req_loc = msg.request(redirect_uri)
                resp = _entity.server.http_request(_req_loc, 'GET')
                ret = Response('Client need to reregister')
                return ret

        self.handle_request(req, path)
        self.store_result()

        self.conv.index += 1

        try:
            resp = self.run_item(self.conv.test_id, index=self.conv.index,
                                 **kwargs)
        except Exception as err:
            raise

        if isinstance(resp, Response):
            self.store_result()
            return resp

        _done = False
        for _cond in self.conv.events.get_data(EV_CONDITION):
            if _cond.test_id == 'Done' and _cond.status == OK:
                _done = True
                break

        if not _done:
            self.conv.events.store(EV_CONDITION, State('Done', OK),
                                   sender='do_next')

            if 'assert' in self.conv.flow:
                _ver = Verify(self.chk_factory, self.conv)
                _ver.test_sequence(self.conv.flow["assert"])

            self.store_result()

        return self.inut.flow_list()

    def get_response(self, resp):
        try:
            loc = resp.headers['location']
        except (AttributeError, KeyError):  # May be a dictionary
            try:
                return resp.response
            except AttributeError:
                try:
                    return resp.text
                except AttributeError:
                    if isinstance(resp, dict):
                        return resp
        else:
            try:
                _resp = dict(
                    [(k, v[0]) for k, v in parse_qs(loc.split('?')[1]).items()])
            except IndexError:
                return loc
            else:
                self.conv.events.store(EV_RESPONSE, _resp)
                return _resp



import importlib
import json
import logging
import sys

from mako.lookup import TemplateLookup
from oic.utils.keyio import build_keyjar

from otest.prof_util import ProfileHandler
from otest.rp.setup import read_path2port_map

from otest.flow import FlowState

logger = logging.getLogger(__name__)

RP_ORDER = [
    "rp-discovery", "rp-registration", "rp-response_type",
    "rp-response_mode", "rp-token_endpoint", "rp-id_token",
    "rp-claims_request", "rp-request_uri", "rp-scope", "rp-nonce",
    "rp-key-rotation", "rp-userinfo", "rp-self-issued", "rp-claims"]

# OP_ORDER = [
#     "OP-Response",
#     "OP-IDToken",
#     "OP-UserInfo",
#     "OP-nonce",
#     "OP-scope",
#     "OP-display",
#     "OP-prompt",
#     "OP-Req",
#     "OP-OAuth",
#     "OP-redirect_uri",
#     "OP-ClientAuth",
#     "OP-Discovery",
#     "OP-Registration",
#     "OP-Rotation",
#     "OP-request_uri",
#     "OP-request",
#     "OP-claims"
# ]

OP_ORDER = [
    "Response Type & Response Mode",
    "Response Type",
    "Response Mode",
    "Discovery",
    "Dynamic Client Registration",
    "redirect_uri",
    "ID Token",
    "Client Authentication",
    'Access Token',
    "Userinfo Endpoint",
    'claims Request Parameter',
    "display Request Parameter",
    "nonce Request Parameter",
    "prompt Request Parameter",
    'redirect_uri Request Parameter',
    "request Request Parameter",
    "request_uri Request Parameter",
    "scope Request Parameter",
    "Misc Request Parameters",
    "OAuth behaviors",
    "Key Rotation",
    "End Session",
    "Session Management",
    "Back Channel Logout - RP Initiated",
    "Front Channel Logout - RP Initiated",
    "Session management - RP Initiated Logout"
]


def construct_app_args(args, conf, operations, func, default_profiles,
                       inst_conf, display_order=None):
    """

    :param args: Command arguments, argparse instance
    :param conf: Service configuration
    :param operations: Operations module
    :param func: Functions module
    :param default_profiles: The default profiles module
    :param inst_conf: Test instance configuration
    :return: Application arguments
    """
    sys.path.insert(0, ".")

    # setup_logging("%s/rp_%s.log" % (SERVER_LOG_FOLDER, _port), logger)

    if args.flowdir:
        _flowdir = args.flowdir
    else:
        _flowdir = conf.FLOWDIR

    cls_factories = {'': operations.factory}
    func_factory = func.factory

    try:
        profiles = importlib.import_module(conf.PROFILES)
    except AttributeError:
        profiles = default_profiles

    if display_order is None:
        display_order = OP_ORDER

    flow_state = FlowState(_flowdir, profile_handler=ProfileHandler,
                           cls_factories=cls_factories,
                           func_factory=func_factory,
                           display_order=display_order)

    # Add own keys for signing/encrypting JWTs
    jwks, keyjar, kidd = build_keyjar(conf.KEYS)

    try:
        if args.staticdir:
            _sdir = args.staticdir
        else:
            _sdir = 'jwks'
    except AttributeError:
        _sdir = 'jwks'

    # If this instance is behind a reverse proxy or on its own
    _port = args.port
    if conf.BASE.endswith('/'):
        conf.BASE = conf.BASE[:-1]

    if args.path2port:
        ppmap = read_path2port_map(args.path2port)
        try:
            _path = ppmap[str(_port)]
        except KeyError:
            print('Port not in path2port map file {}'.format(args.path2port))
            sys.exit(-1)

        # if args.xport:
        #     _base = '{}:{}/{}/'.format(conf.BASE, str(_port), _path)
        # else:
        _base = '{}/{}/'.format(conf.BASE, _path)
    else:
        if _port not in [443, 80]:
            _base = '{}:{}'.format(conf.BASE, _port)
        else:
            _base = conf.BASE
        _path = ''

    if not _base.endswith('/'):
        _base += '/'

    # -------- JWKS ---------------

    if args.path2port:
        jwks_uri = "{}{}/jwks_{}.json".format(_base, _sdir, _port)
        f = open('{}/jwks_{}.json'.format(_sdir, _port), "w")
    elif _port not in [443, 80]:
        jwks_uri = "{}:{}/{}/jwks_{}.json".format(conf.BASE, _port, _sdir,
                                                  _port)
        f = open('{}/jwks_{}.json'.format(_sdir, _port), "w")
    else:
        jwks_uri = "{}/{}/jwks.json".format(conf.BASE, _sdir)
        f = open('{}/jwks.json'.format(_sdir), "w")
    f.write(json.dumps(jwks))
    f.close()

    # -------- MAKO setup -----------
    try:
        if args.makodir:
            _dir = args.makodir
            if not _dir.endswith("/"):
                _dir += "/"
        else:
            _dir = "./"
    except AttributeError:
        _dir = './'

    LOOKUP = TemplateLookup(directories=[_dir + 'templates', _dir + 'htdocs'],
                            module_directory=_dir + 'modules',
                            input_encoding='utf-8',
                            output_encoding='utf-8')

    _client_info = inst_conf['client']

    # Now when the basic URL for the RP is constructed update the
    # redirect_uris and the post_logout_redirect_uris
    try:
        ri = _client_info['registration_info']
    except KeyError:
        pass
    else:
        ri['redirect_uris'] = [r.format(_base) for r in ri['redirect_uris']]
        try:
            ri['post_logout_redirect_uris'] = [r.format(_base) for r in
                                               ri['post_logout_redirect_uris']]
        except KeyError:
            pass

    _client_info.update(
        {"base_url": _base, "kid": kidd, "keyjar": keyjar,
         "jwks_uri": jwks_uri}
    )

    # try:
    #     _client_info['client_id'] = _client_info['registration_response'][
    #         'client_id']
    # except KeyError:
    #     pass

    if args.insecure:
        _client_info['verify_ssl'] = False

    # Test profile either as a command line argument or if not that
    # from the configuration file
    # if args.profile:
    #     _profile = args.profile
    # else:
    # _profile = inst_conf['tool']['profile']

    # Application arguments
    app_args = {
        "flow_state": flow_state, "conf": conf, "base_url": _base,
        "client_info": _client_info, "profiles": profiles,
        "operation": operations, "cache": {},  # "profile": _profile,
        "lookup": LOOKUP, 'tool_conf': inst_conf['tool'],
        "profile_handler": ProfileHandler
    }

    return _path, app_args



import logging

from otest import ConditionError
from otest import Done
from otest import exception_trace
from otest.check import CRITICAL
from otest.check import ERROR
from otest.check import OK
from otest.check import State
from otest.conversation import Conversation
from otest.events import EV_CONDITION
from otest.events import EV_FAULT
from otest.result import Result
from otest.verify import Verify

__author__ = 'roland'

logger = logging.getLogger(__name__)


def get_redirect_uris(cinfo):
    """
    Used before there is a Conversation instance
    :param cinfo: Client Configuration Information
    :return: list of redirect_uris
    """
    try:
        res = cinfo["registration_info"]["redirect_uris"]
    except KeyError:
        res = cinfo["registration_response"]["redirect_uris"]

    if isinstance(res, ("".__class__, u"".__class__)):
        return [res]
    else:
        return res


class ConfigurationError(Exception):
    pass


class Tester(object):
    def __init__(self, inut, sh, profile='', flows=None, check_factory=None,
                 msg_factory=None, cache=None, make_entity=None, map_prof=None,
                 com_handler=None, response_cls=None,
                 client_factory=None, **kwargs):
        self.inut = inut
        self.sh = sh
        self.conv = None
        self.profile = profile
        self.flows = flows
        self.message_factory = msg_factory
        self.check_factory = check_factory
        self.client_factory = client_factory
        self.cache = cache
        self.kwargs = kwargs
        self.make_entity = make_entity
        self.map_prof = map_prof
        self.com_handler = com_handler
        self.cjar = {}
        self.response_cls = response_cls

    def match_profile(self, test_id):
        _spec = self.flows[test_id]
        return self.map_prof(self.profile.split("."),
                             _spec["profile"].split("."))

    def setup(self, test_id, **kw_args):
        redirs = get_redirect_uris(kw_args['client_info'])

        _flow = self.flows[test_id]
        _cli, _cli_info = self.client_factory.make_client(
            **kw_args['client_info'])
        self.conv = Conversation(_flow, _cli, kw_args["msg_factory"],
                                 callback_uris=redirs)
        self.conv.entity_config = _cli_info
        self.conv.tool_config = kw_args['tool_conf']
        _cli.conv = self.conv
        _cli.events = self.conv.events
        self.sh.session_setup(path=test_id)
        self.sh["conv"] = self.conv
        self.conv.sequence = self.sh["sequence"]
        return True

    def run(self, test_id, **kw_args):
        if not self.setup(test_id, **kw_args):
            raise ConfigurationError()

        # noinspection PyTypeChecker
        try:
            return self.run_flow(test_id, conf=kw_args['conf'])
        except Exception as err:
            exception_trace("", err, logger)
            self.store_result()
            return self.inut.err_response("run", err)

    def handle_response(self, resp, index, oper=None):
        return None

    def fname(self, test_id):
        raise NotImplemented()

    def get_response(self, resp):
        try:
            return resp.response
        except AttributeError:
            return resp.text

    def store_result(self, res=None):
        tinfo = self.flows.store_test_info(self)
        if res is None:
            res = Result(self.sh, self.kwargs['profile_handler'])
        res.write_info(tinfo)
        return tinfo

    def run_flow(self, test_id, index=0, profiles=None, **kwargs):
        logger.info("<=<=<=<=< %s >=>=>=>=>" % test_id)
        _ss = self.sh
        _ss.test_flows.complete[test_id] = False

        self.conv.test_id = test_id
        res = Result(self.sh, self.kwargs['profile_handler'])

        if index >= len(self.conv.sequence):
            return CRITICAL

        _oper = None
        for item in self.conv.sequence[index:]:
            if isinstance(item, tuple):
                cls, funcs = item
            else:
                cls = item
                funcs = {}

            logger.info("<--<-- {} --- {} -->-->".format(index, cls))
            self.conv.events.store('operation', cls, sender='run_flow')
            try:
                _oper = cls(conv=self.conv, inut=self.inut, sh=self.sh,
                            profile=self.profile, test_id=test_id,
                            funcs=funcs, check_factory=self.check_factory,
                            cache=self.cache)
                # self.conv.operation = _oper
                if profiles:
                    profile_map = profiles.PROFILEMAP
                else:
                    profile_map = None
                _oper.setup(profile_map)
                resp = _oper()
            except ConditionError:
                self.store_result(res)
                return ERROR
            except Exception as err:
                exception_trace('run_flow', err)
                self.conv.events.store(EV_FAULT, err)
                #self.sh["index"] = index
                self.store_result(res)
                return CRITICAL
            else:
                if isinstance(resp, self.response_cls):
                    return resp

                if resp:
                    if self.com_handler:
                        resp = self.com_handler(resp)

                    resp = _oper.handle_response(self.get_response(resp))

                    if resp:
                        return self.inut.respond(resp)

            # should be done as late as possible, so all processing has been
            # done
            try:
                _oper.post_tests()
            except ConditionError:
                tinfo = self.store_result(res)
                return tinfo['state']

            index += 1

        _ss['index'] = self.conv.index = index

        try:
            if self.conv.flow["assert"]:
                _ver = Verify(self.check_factory, self.conv)
                _ver.test_sequence(self.conv.flow["assert"])
        except KeyError:
            pass
        except Exception as err:
            logger.error(err)
            raise

        if isinstance(_oper, Done):
            self.conv.events.store(EV_CONDITION, State('Done', OK),
                                   sender='run_flow')

        tinfo = self.store_result(res)

        return tinfo['state']



import json
import yaml
from otest import Unknown
from otest.func import factory as ofactory

__author__ = 'roland'


class MissingParent(Exception):
    pass


def _get_cls(name, factories, use=''):
    if use:
        try:
            cls = factories[use](name)
        except Unknown:
            pass
        else:
            return cls

    try:
        cls = factories[''](name)
    except Unknown:
        raise Exception("Unknown Class: '{}'".format(name))

    return cls


def _get_func(dic, func_factory):
    """
    Convert function names into function references

    :param dic: A key, value dictionary where keys are function names
    :param func_factory: Factory function used to find functions
    :return: A dictionary with the keys replace with references to functions
    """
    res = {}
    for fname, val in dic.items():
        func = func_factory(fname)
        if func is None:
            func = ofactory(fname)

        if func is None:
            raise Exception("Unknown function: '{}'".format(fname))
        res[func] = val

    return res


def parse_yaml_conf(cnf_file, cls_factories, func_factory, use=''):
    """

    :param cnf_file:
    :param use:
    :return:
    """
    stream = open(cnf_file, 'r')
    yc = yaml.safe_load(stream)
    stream.close()
    for tid, spec in yc['Flows'].items():
        seq = []
        for oper in spec["sequence"]:
            if isinstance(oper, dict):  # Must be only one key, value item
                if len(oper) > 1:
                    raise SyntaxError(tid)
                key, val = list(oper.items())[0]
                try:
                    seq.append((_get_cls(key, cls_factories, use),
                                _get_func(val, func_factory)))
                except Exception:
                    print('tid:{}'.format(tid))
                    raise
            else:
                try:
                    seq.append(_get_cls(oper, cls_factories, use))
                except Exception:
                    print('tid:{}'.format(tid))
                    raise
        spec["sequence"] = seq

    return yc


def parse_json_conf(cnf_file, cls_factories, func_factory, use=''):
    """

    :param cnf_file:
    :param use:
    :return:
    """
    stream = open(cnf_file, 'r')
    js = json.load(stream)
    stream.close()
    for tid, spec in js['Flows'].items():
        seq = []
        for oper in spec["sequence"]:
            if isinstance(oper, dict):  # Must be only one key, value item
                if len(oper) > 1:
                    raise SyntaxError(tid)
                key, val = list(oper.items())[0]
                try:
                    _cls = _get_cls(key, cls_factories, use)
                except Exception:
                    print('Failed to get class on tid:{}'.format(tid))
                    raise
                else:
                    try:
                        _func = _get_func(val, func_factory)
                    except Exception:
                        print('Failed to get function on tid:{}'.format(tid))
                        raise
                    else:
                        seq.append((_cls,_func))
            else:
                try:
                    seq.append(_get_cls(oper, cls_factories, use))
                except Exception:
                    print('Failed to get class on tid:{}'.format(tid))
                    raise
        spec["sequence"] = seq

    return js


class Item(object):
    def __init__(self, parent, name, desc):
        self.parent = parent
        self.desc = desc
        self.name = name
        self.child = []


def build_hierarchy(flows):
    items = {}

    for id, desc in flows.items():
        items[id] = Item('', id, desc)

    for item in items.values():
        try:
            _pre = item.desc['super']
        except KeyError:
            continue
        else:
            try:
                _parent = items[_pre]
                _parent.child.append(item)
                item.parent = _parent
            except KeyError:
                raise MissingParent(item.desc['super'])

    return items


def flatten(interim):
    res = []
    for f in interim:
        res.append(f)
        if f.child:
            res.extend(flatten(sorted(f.child, key=lambda x: x.name)))
    return res


def sort(display_order, flows):
    items = build_hierarchy(flows)
    # toplevel
    f_names = [f for f in items.values() if not f.parent]
    interim = []
    for k in display_order:
        k += '-'
        l = [z for z in f_names if z.name.startswith(k)]
        interim.extend(sorted(l, key=lambda x: x.name))

    return flatten(interim)



import copy
import inspect
import logging
import sys

from bs4 import BeautifulSoup
from oic.exception import IssuerMismatch
from oic.oauth2 import ResponseError
from oic.oauth2.message import AccessTokenResponse
from oic.oauth2.message import ErrorResponse
from oic.oauth2.message import Message
from oic.oauth2.message import MissingRequiredAttribute
from oic.oauth2.message import VerificationError
from oic.oauth2.util import URL_ENCODED
from oic.oic.message import IdToken
from oic.utils.http_util import Redirect
from requests.models import Response

from otest import Break
from otest import Unknown
from otest import operation
from otest.check import ERROR
from otest.check import State
from otest.events import EV_FAULT
from otest.events import EV_HTTP_RESPONSE
from otest.events import EV_JWE_HEADER
from otest.events import EV_JWS_HEADER
from otest.events import EV_PROTOCOL_RESPONSE
from otest.events import EV_REDIRECT_URL
from otest.events import EV_REQUEST
from otest.events import EV_RESPONSE
from otest.prof_util import RESPONSE

__author__ = 'rolandh'

logger = logging.getLogger(__name__)

DUMMY_URL = "https://remove.this.url/"


class ParameterError(Exception):
    pass


class Operation(operation.Operation):
    message_cls = Message


class Request(Operation):
    def expected_error_response(self, response):
        if isinstance(response, Response):  # requests response
            # don't want bytes
            _txt = response.content.decode('utf8')
            response = ErrorResponse().from_json(_txt)

        if isinstance(response, ErrorResponse):
            self.conv.events.store(EV_PROTOCOL_RESPONSE, response,
                                   sender=self.__class__.__name__)
            if response["error"] not in self.expect_error["error"]:
                raise Break("Wrong error, got {} expected {}".format(
                    response["error"], self.expect_error["error"]))
            try:
                if self.expect_error["stop"]:
                    raise Break("Stop requested after received expected error")
            except KeyError:
                pass
        else:
            self.conv.events.store(EV_FAULT, "Expected error, didn't get it")
            raise Break("Did not receive expected error")

        return response

    def map_profile(self, profile_map):
        if '.' in self.profile:
            _rt = self.profile.split('.')[RESPONSE]
        else:
            _rt = self.profile[RESPONSE]

        try:
            item = profile_map[self.__class__][_rt]
        except KeyError:
            pass
        else:
            for func, arg in item.items():
                func(self, arg)


class SyncRequest(Request):
    request_cls = None
    method = ""
    module = ""
    content_type = URL_ENCODED
    response_cls = None
    response_where = "url"
    response_type = "urlencoded"
    accept = None
    _tests = {"post": [], "pre": []}

    def __init__(self, conv, inut, sh, **kwargs):
        Operation.__init__(self, conv, inut, sh, **kwargs)

        try:
            self.profile = self.profile.split('.')
        except AttributeError:
            pass

        self.conv.req = self
        self.tests = copy.deepcopy(self._tests)
        if self.request_cls:
            self.request = self.conv.msg_factory(self.request_cls)
        else:
            self.request = Message
        if self.response_cls:
            self.response = self.conv.msg_factory(self.response_cls)
        else:
            self.response = Message

    def do_request(self, client, url, body, ht_args):
        response = client.http_request(url, method=self.method, data=body,
                                       **ht_args)

        self.conv.events.store(EV_HTTP_RESPONSE, response,
                               sender=self.__class__.__name__)

        return response

    def handle_response(self, r, csi):
        data = self.conv.events.last_item(EV_REQUEST)
        try:
            state = data['state']
        except KeyError:
            state = ''

        if 300 < r.status_code < 400:
            resp = self.conv.entity.parse_response(
                self.response, info=r.headers['location'],
                sformat="urlencoded", state=state)
        elif r.status_code == 200:
            if "response_mode" in csi and csi["response_mode"] == "form_post":
                resp = self.response()
                forms = BeautifulSoup(r.text).findAll('form')
                for inp in forms[0].find_all("input"):
                    resp[inp.attrs["name"]] = inp.attrs["value"]
            else:
                if r.is_redirect or r.is_permanent_redirect:
                    resp = self.conv.entity.parse_response(
                        self.response, info=r.headers['location'],
                        sformat="urlencoded", state=state)
                else:
                    resp = self.conv.entity.parse_response(
                        self.response, info=r.text,
                        sformat="json", state=state)

            _ent = self.conv.entity
            if isinstance(resp, AccessTokenResponse):
                if 'id_token' in resp and isinstance(resp['id_token'], IdToken):
                    pass
                else:
                    resp.verify(keyjar=_ent.keyjar, client_id=_ent.client_id,
                                iss=_ent.provider_info['issuer'])
            else:
                resp.verify(keyjar=_ent.keyjar, client_id=_ent.client_id,
                            iss=_ent.provider_info['issuer'])

        elif r.status_code == 400:
            if r.headers['content-type'] == 'application/json':
                resp = ErrorResponse().from_json(r.text)
            else:
                resp = ErrorResponse(error='service_error',
                                     error_description=r.text)
        else:
            resp = r

        if not isinstance(resp, Response):  # Not a HTTP response
            try:
                try:
                    _id_token = resp["id_token"]
                except KeyError:
                    pass
                else:
                    if _id_token.jws_header:
                        self.conv.events.store('JWS header',
                                               _id_token.jws_header)
                    if _id_token.jwe_header:
                        self.conv.events.store('JWE header',
                                               _id_token.jwe_header)
                    if "kid" not in _id_token.jws_header and not \
                            _id_token.jws_header["alg"] == "HS256":
                        keys = self.conv.entity.keyjar.keys_by_alg_and_usage(
                            issuer=_id_token['iss'],
                            alg=_id_token.jws_header["alg"],
                            usage='sig'
                        )
                        if len(keys) > 1:
                            raise ParameterError("No 'kid' in id_token header!")

                    try:
                        if self.req_args['nonce'] != _id_token['nonce']:
                            raise ParameterError(
                                "invalid nonce! {} != {}".format(
                                    self.req_args['nonce'], _id_token['nonce']))
                    except KeyError:
                        pass

                    if not same_issuer(self.conv.info["issuer"],
                                       _id_token["iss"]):
                        raise IssuerMismatch(
                            " {} != {}".format(self.conv.info["issuer"],
                                               _id_token["iss"]))
                    self.conv.entity.id_token = _id_token
            except KeyError:
                pass

        return resp

    def run(self):
        _client = self.conv.entity

        url, body, ht_args, csi = _client.request_info(
            self.request, method=self.method, request_args=self.req_args,
            target=_client.provider_info["issuer"],
            **self.op_args)

        try:
            http_args = self.op_args["http_args"]
        except KeyError:
            http_args = ht_args
        else:
            http_args.update(ht_args)

        self.conv.events.store(EV_REQUEST, csi, sender=self.__class__.__name__)
        http_response = self.do_request(_client, url, body, http_args)

        self.catch_exception_and_error(self.handle_response, r=http_response,
                                       csi=csi)


class AsyncRequest(Request):
    request_cls = None
    method = ""
    module = ""
    content_type = URL_ENCODED
    response_cls = ""
    response_where = "url"  # if code otherwise 'body'
    response_type = "urlencoded"
    accept = None
    _tests = {"post": [], "pre": []}

    def __init__(self, conv, inut, sh, **kwargs):
        Operation.__init__(self, conv, inut, sh, **kwargs)
        try:
            self.profile = self.profile.split('.')
        except AttributeError:
            pass
        self.conv.req = self
        self.tests = copy.deepcopy(self._tests)
        self.csi = None
        self.request = self.conv.msg_factory(self.request_cls)
        self.response = self.conv.msg_factory(self.response_cls)

    def run(self):
        _client = self.conv.entity

        url, body, ht_args, csi = _client.request_info(
            self.request, method=self.method, request_args=self.req_args,
            lax=True, **self.op_args)

        self.csi = csi

        self.conv.events.store(EV_REDIRECT_URL, url,
                               sender=self.__class__.__name__)
        return Redirect(str(url))

    def parse_response(self, path, inut, message_factory, response=None):
        _ctype = self.response_type
        _conv = self.conv

        if self.csi is None:
            url, body, ht_args, csi = _conv.entity.request_info(
                self.request, method=self.method, request_args=self.req_args,
                **self.op_args)

            self.csi = csi

        try:
            response_mode = self.csi["response_mode"]
        except KeyError:
            response_mode = None

        if self.request_cls == "AuthorizationRequest":
            try:
                _rt = self.csi["response_type"]
            except KeyError:
                response_where = ""
            else:
                if _rt == ["code"]:
                    response_where = "url"
                elif _rt == [""]:
                    response_where = ""
                else:
                    response_where = "fragment"
        else:
            response_where = self.response_where

        # parse the response
        if response_mode == "form_post":
            info = response
            _ctype = "dict"
        elif response_where in ["url", ""]:
            info = response
            _ctype = "dict"
        elif response_where == "fragment":
            try:
                info = response["fragment"]
            except KeyError:
                return inut.sorry_response(inut.base_url, "missing fragment ?!")
        else:  # resp_c.where == "body"
            info = response

        logger.info("Response: %s" % info)

        ev_index = _conv.events.store(EV_RESPONSE, info,
                                      sender=self.__class__.__name__)

        resp_cls = message_factory(self.response_cls)
        #  algs = _conv.entity.sign_enc_algs("id_token")
        try:
            response = _conv.entity.parse_response(
                resp_cls, info, _ctype,
                self.csi["state"],
                keyjar=_conv.entity.keyjar  # , algs=algs
            )
        except ResponseError as err:
            _conv.events.store(EV_FAULT, State(_conv.test_id, ERROR,
                                               message=err,
                                               context='parse_response'))
            return inut.err_response("run_sequence", err)
        except (VerificationError, MissingRequiredAttribute) as err:
            self.conv.events.store(EV_FAULT, err)
            inut.err_response("run_sequence", err)
            return None
        except Exception as err:
            return inut.err_response("run_sequence", err)

        logger.info("Parsed response: %s" % response.to_dict())

        _conv.events.store(EV_PROTOCOL_RESPONSE, response, ref=ev_index,
                           sender=self.__class__.__name__)

        display_jwx_headers(response, _conv)

        if self.expect_error:
            self.expected_error_response(response)
        else:
            if isinstance(response, ErrorResponse):
                raise Break("Unexpected error response")


def display_jwx_headers(message, conv):
    try:
        _jws_header = message.jws_header
    except (KeyError, AttributeError):
        pass
    else:
        if _jws_header:
            conv.events.store(EV_JWS_HEADER, "{}".format(_jws_header))

    try:
        _jwe_header = message.jwe_header
    except KeyError:
        pass
    else:
        if _jwe_header:
            conv.events.store(EV_JWE_HEADER, "{}".format(_jwe_header))


def same_issuer(issuer_A, issuer_B):
    if not issuer_A.endswith("/"):
        issuer_A += "/"
    if not issuer_B.endswith("/"):
        issuer_B += "/"
    return issuer_A == issuer_B


class SyncGetRequest(SyncRequest):
    method = "GET"


class AsyncGetRequest(AsyncRequest):
    method = "GET"


class SyncPostRequest(SyncRequest):
    method = "POST"


class SyncPutRequest(SyncRequest):
    method = "PUT"


class SyncDeleteRequest(SyncRequest):
    method = "DELETE"


def factory(name):
    for fname, obj in inspect.getmembers(sys.modules[__name__]):
        if inspect.isclass(obj):
            if name == fname:
                return obj

    obj = operation.factory(name)
    if not obj:
        raise Unknown("Couldn't find the operation: '{}'".format(name))
    return obj



__author__ = 'roland'


class HandlerResponse(object):
    def __init__(self, content_processed, outside_html_action=None,
                 tester_error_description=None,
                 cookie_jar=None, urllib_request=None, urllib_response=None):
        """

        :param content_processed: bool set to True if a scripted ContentHandler
        matches and processes a page; If False then the next ContentHandler
        must take over
        :param cookie_jar: A CookieJar instance
        :param urllib_response: A urllib.response.addinfourl instance
        :param outside_html_action: Value from outside_html_actions or None
        :param tester_error_description: optional text if outside_html_action
        is not None
        :param response: A semi parsed response, might be a dictionary
        """
        self.content_processed = content_processed
        self.outside_html_action = outside_html_action
        self.cookie_jar = cookie_jar
        self.urllib_request = urllib_request
        self.urllib_response = urllib_response
        self.tester_error_description = tester_error_description


class ContentHandler(object):
    """
    Process the HTML contents of a response from the test target. This can
    either be a scripted approach, or invoke a browser.
    """

    def __init__(self):
        pass

    def handle_response(self, http_response, auto_close_urls,
                        conv=None, verify_ssl=True, cookie_jar=None,
                        outside_html_actions=None):
        """

        :param http_response: The HTTP response to handle
        :param auto_close_urls: A list of URLs that if encountered should
        lead to an immediate break in processing, like a form action. Other
        URLs in the page will load local resources such as css and js without
        returning control.
        :param conv: A otest.Conversation instance
        :param verify_ssl: (True/False) whether the ssl certificates must
        be verified. Default is True
        :param cookie_jar: A http.cookiejar.CookieJar instance
        :param outside_html_actions: a dict describing buttons for the widget
        outside the html-area, to be used if the test must be aborted
        :return: A otest.contenthandler.HandlerResponse instance
        """
        raise NotImplemented()



import logging
import mimetypes
import os

from otest.check import State
from otest.check import ERROR
from otest.events import EV_CONDITION
from otest.events import EV_HTTP_INFO
from otest.events import EV_REQUEST

from future.backports.urllib.parse import parse_qs

from oic.utils.http_util import BadRequest
from oic.utils.http_util import extract_from_request
from oic.utils.http_util import NotFound
from oic.utils.http_util import Response
from oic.utils.http_util import ServiceError

from oic.utils.webfinger import OIC_ISSUER
from oic.utils.webfinger import WebFinger

__author__ = 'roland'

logger = logging.getLogger(__name__)


def store_response(response, events):
    events.store(EV_HTTP_INFO, response.info())


def wsgi_wrapper(environ, func, events, **kwargs):
    kwargs = extract_from_request(environ, kwargs)
    if kwargs['request']:
        events.store(EV_REQUEST, kwargs['request'])
    args = func(**kwargs)

    try:
        resp, state = args
        store_response(resp, events)
        return resp
    except TypeError:
        resp = args
        store_response(resp, events)
        return resp
    except Exception as err:
        logger.error("%s" % err)
        raise


# noinspection PyUnresolvedReferences
def static_mime(path, environ, start_response):
    logger.info("[static]sending: %s" % (path,))

    # Set content-type based on filename extension
    ext = ""
    i = path.rfind('.')
    if i != -1:
        ext = path[i:].lower()
    content_type = mimetypes.types_map.get(ext, None)

    try:
        if not content_type.startswith('image/'):
            data = open(path, 'r').read()
        else:
            data = open(path, 'rb').read()
        resp = Response(data, content=content_type)
        return resp(environ, start_response)
    except IOError:
        _dir = os.getcwd()
        resp = NotFound("{} not in {}".format(path, _dir))
    except Exception as err:
        resp = NotFound('{}'.format(err))

    return resp(environ, start_response)


# noinspection PyUnresolvedReferences
def static(path):
    logger.info("[static]sending: %s" % (path,))

    try:
        resp = Response(open(path).read())
        if path.endswith(".ico"):
            resp.add_header(('Content-Type', "image/x-icon"))
        elif path.endswith(".html"):
            resp.add_header(('Content-Type', 'text/html'))
        elif path.endswith(".json"):
            resp.add_header(('Content-Type', 'application/json'))
        elif path.endswith(".txt"):
            resp.add_header(('Content-Type', 'text/plain'))
        elif path.endswith(".css"):
            resp.add_header(('Content-Type', 'text/css'))
        else:
            resp.add_header(('Content-Type', "text/xml"))
        return resp
    except IOError:
        return NotFound(path)


def css(environ, events):
    try:
        info = open(environ["PATH_INFO"]).read()
        resp = Response(info)
    except (OSError, IOError):
        resp = NotFound(environ["PATH_INFO"])

    return resp


def token(environ, events):
    _op = environ["oic.op"]

    return wsgi_wrapper(environ, _op.token_endpoint, events)


def authorization(environ, events):
    _op = environ["oic.op"]

    return wsgi_wrapper(environ, _op.authorization_endpoint,
                        events)


def userinfo(environ, events):
    _op = environ["oic.op"]

    return wsgi_wrapper(environ, _op.userinfo_endpoint,
                        events)


def clientinfo(environ, events):
    _op = environ["oic.op"]

    return wsgi_wrapper(environ, _op.client_info_endpoint,
                        events)


def revocation(environ, events):
    _op = environ["oic.op"]

    return wsgi_wrapper(environ, _op.revocation_endpoint,
                        events)


def introspection(environ, events):
    _op = environ["oic.op"]

    return wsgi_wrapper(environ, _op.introspection_endpoint, events)


# noinspection PyUnusedLocal
def op_info(environ, events):
    _op = environ["oic.op"]
    logger.info("op_info")
    return wsgi_wrapper(environ, _op.providerinfo_endpoint,
                        events)


# noinspection PyUnusedLocal
def registration(environ, events):
    _op = environ["oic.op"]

    if environ["REQUEST_METHOD"] == "POST":
        return wsgi_wrapper(environ, _op.registration_endpoint,
                            events)
    elif environ["REQUEST_METHOD"] == "GET":
        return wsgi_wrapper(environ, _op.read_registration,
                            events)
    else:
        return ServiceError("Method not supported")


def webfinger(environ, events):
    query = parse_qs(environ["QUERY_STRING"])
    _op = environ["oic.op"]

    try:
        if query["rel"] != [OIC_ISSUER]:
            events.store(
                EV_CONDITION,
                State('webfinger_parameters', ERROR,
                      message='parameter rel wrong value: {}'.format(
                          query['rel'])))
            return BadRequest('Parameter value error')
        else:
            resource = query["resource"][0]
    except KeyError as err:
        events.store(EV_CONDITION,
                       State('webfinger_parameters', ERROR,
                             message='parameter {} missing'.format(err)))
        resp = BadRequest("Missing parameter in request")
    else:
        wf = WebFinger()
        resp = Response(wf.response(subject=resource, base=_op.baseurl))
    return resp


def add_endpoints(extra, URLS):
    for endp in extra:
        URLS.append(("^%s" % endp.etype, endp.func))

    return URLS



import logging
from otest import FatalError
from otest.contenthandler import HandlerResponse
from otest.events import EV_HTTP_RESPONSE

__author__ = 'roland'

logger = logging.getLogger(__name__)


class ComHandler(object):
    def __init__(self, contenthandlers, conv=None, auto_close_urls=None):
        self.content_handlers = contenthandlers
        self.conv = conv
        self.auto_close_urls = auto_close_urls or []
        self.verify_ssl = True

    def __call__(self, http_response, target_url='', auto_close_urls=None,
                 conv=None, **kwargs):
        if not http_response:
            return

        auto_close_urls = auto_close_urls or []
        auto_close_urls.extend(self.auto_close_urls)
        rdseq = []

        while True:
            if http_response.status_code >= 400:
                return http_response

            while http_response.status_code in [300,301,302]:
                url = http_response.headers["location"]
                if url in rdseq:
                    raise FatalError("Loop detected in redirects")
                else:
                    rdseq.append(url)
                    if len(rdseq) > 8:
                        raise FatalError(
                            "Too long sequence of redirects: %s" % rdseq)

                logger.info("HTTP %d Location: %s" % (http_response.status_code,
                                                      url))

                if url in auto_close_urls:
                    return http_response

                try:
                    logger.info("GET %s" % url)
                    http_response = self.conv.entity.send(url, "GET")
                except Exception as err:
                    raise FatalError("%s" % err)

                self.conv.events(EV_HTTP_RESPONSE, http_response)

                if http_response.status_code >= 400:
                    return http_response

            handled = False
            for ct in self.content_handlers:
                resp = ct.handle_response(http_response, auto_close_urls,
                                          conv=self.conv,
                                          verify_ssl=self.verify_ssl,
                                          cookie_jar=self.conv.entity.cookiejar)
                if resp.content_processed:
                    if resp.cookie_jar:
                        self.conv.entity.cookie_jar = resp.cookie_jar
                    if resp.http_response:
                        http_response = resp.http_response
                    else:
                        return resp
                    handled = True
                    break

            if not handled:
                return HandlerResponse(False)



import copy
import inspect
import logging
import sys

from future.backports.http.cookies import CookieError
from future.backports.http.cookies import SimpleCookie
from future.backports.urllib.parse import urlparse
from oic import oic
from oic.extension.oidc_fed import ProviderConfigurationResponse
from oic.oauth2 import ErrorResponse
from oic.utils.http_util import SeeOther

from otest import operation
from otest import OperationError

from otest.events import EV_HTTP_RESPONSE
from otest.events import EV_RESPONSE
from otest.events import OUTGOING
from otest.events import NoSuchEvent
from otest.events import EV_RUN
from otest.events import EV_PROTOCOL_RESPONSE
from otest.events import EV_REDIRECT_URL
from otest.events import EV_REQUEST
from otest.rp.response import Response
from otest.verify import Verify

from oic.oauth2.util import set_cookie

from oic.oic import message

__author__ = 'roland'

logger = logging.getLogger(__name__)


def save_response(events, resp, resp_cls, err_resp_cls):
    if 'error' in resp.message:
        _msgtype = err_resp_cls
    else:
        _msgtype = resp_cls

    if isinstance(resp, SeeOther):
        p = urlparse(resp.message)
        if p.query:
            _msg = _msgtype().from_urlencoded(p.query)
        elif p.fragment:
            _msg = _msgtype().from_urlencoded(p.query)
        else:
            _msg = ''
    else:
        _msg = copy.copy(resp.message)
        try:
            _msg = _msgtype().from_json(_msg)
        except Exception as err:
            _msg = _msgtype().from_urlencoded(_msg)

    events.store(EV_PROTOCOL_RESPONSE, _msg, direction=OUTGOING)
    events.store(EV_RESPONSE, resp.message, direction=OUTGOING)


class Operation(operation.Operation):
    def __call__(self, *args, **kwargs):
        if self.skip:
            return
        else:
            cls_name = self.__class__.__name__
            if self.tests["pre"]:
                _ver = Verify(self.check_factory, self.conv, cls_name=cls_name)
                _ver.test_sequence(self.tests["pre"])

            self.conv.events.store(EV_RUN, cls_name)
            res = self.run(*args, **kwargs)

            if res:
                return res


class Init(Operation):
    start_page = ''
    endpoint = ''

    def run(self, **kwargs):
        self.conv.events.store('start_page', self.start_page)
        if self.internal:
            res = self.conv.entity.server.http_request(self.start_page)
        else:
            res = SeeOther(self.start_page)
            return res
        self.conv.events.store(EV_HTTP_RESPONSE, res)
        if res.status_code in [302, 303]:
            loc = res.headers['location']
            try:
                self.conv.events.store('Cookie', res.headers['set-cookie'])
            except KeyError:
                pass
            logger.info('Redirect to {}'.format(loc))
            logger.debug('msg: {}'.format(res.text))
            self.conv.events.store(EV_REDIRECT_URL, loc, sub='init')
        elif res.status_code >= 400:
            logger.info('Error {}'.format(res.text))
            raise OperationError('Error response on HTTP request')
        return res

    def handle_response(self, resp, *args):
        self.conv.events.store(EV_PROTOCOL_RESPONSE,
                               message.AuthorizationResponse(**resp))


class ConfigurationResponse(Response):
    endpoint = 'providerinfo_endpoint'

    def __init__(self, conv, inut, sh, **kwargs):
        Response.__init__(self, conv, inut, sh, **kwargs)
        try:
            self.op_type = kwargs['op_type']
        except KeyError:
            self.op_type = ''
        else:
            del kwargs['op_type']
        self.msg_args = kwargs

    def handle_request(self, *args):
        return None

    def construct_message(self):
        op = self.conv.entity
        resp = op.providerinfo_endpoint()
        if resp.status == '200 OK' or resp.status == '201 Created':
            save_response(self.conv.events, resp, ProviderConfigurationResponse,
                          ErrorResponse)
        return resp


class RegistrationResponse(Response):
    endpoint = 'registration'

    def __init__(self, conv, inut, sh, **kwargs):
        Response.__init__(self, conv, inut, sh, **kwargs)
        try:
            self.op_type = kwargs['op_type']
        except KeyError:
            self.op_type = ''
        else:
            del kwargs['op_type']
        self.msg_args = kwargs

    def construct_message(self):
        req = self.conv.events.last_item(EV_REQUEST)
        resp = self.conv.entity.registration_endpoint(req)
        if resp.status == '200 OK' or resp.status == '201 Created':
            logging.debug('Registration response: {}'.format(resp.message))
            save_response(self.conv.events, resp,
                          oic.message.RegistrationResponse,
                          oic.message.ClientRegistrationErrorResponse)
        return resp


class AuthorizationResponse(Response):
    endpoint = 'authorization'

    def __init__(self, conv, inut, sh, **kwargs):
        Response.__init__(self, conv, inut, sh, **kwargs)
        try:
            self.op_type = kwargs['op_type']
        except KeyError:
            self.op_type = ''
        else:
            del kwargs['op_type']
        self.msg_args = kwargs

    def construct_message(self):
        _kwargs = {'request': self.conv.events.last_item(EV_REQUEST)}
        _kwargs.update(self.msg_args)
        _kwargs.update(self.op_args)

        _op = self.conv.entity
        try:
            _cookie = self.conv.events.last_item('Cookie')
        except NoSuchEvent:
            pass
        else:
            try:
                set_cookie(_op.server.cookiejar, SimpleCookie(_cookie))
            except CookieError as err:
                logger.error(err)
            else:
                _kwargs['cookie'] = _op.server._cookies()

        resp = _op.authorization_endpoint(**_kwargs)
        if isinstance(resp, SeeOther):
            save_response(self.conv.events, resp,
                          oic.message.AuthorizationResponse,
                          oic.message.AuthorizationErrorResponse)
        return resp


class AccessTokenResponse(Response):
    endpoint = 'token'

    def __init__(self, conv, inut, sh, **kwargs):
        Response.__init__(self, conv, inut, sh, **kwargs)
        try:
            self.op_type = kwargs['op_type']
        except KeyError:
            self.op_type = ''
        else:
            del kwargs['op_type']
        self.msg_args = kwargs

    def construct_message(self):
        _kwargs = {
            'request': self.conv.events.last_item(EV_REQUEST),
        }

        try:
            _kwargs['authn'] = self.conv.events.last_item('HTTP_AUTHORIZATION')
        except NoSuchEvent:
            pass

        _kwargs.update(self.msg_args)
        _kwargs.update(self.op_args)

        resp = self.conv.entity.token_endpoint(**_kwargs)
        save_response(self.conv.events, resp,
                      oic.message.AccessTokenResponse,
                      oic.message.TokenErrorResponse)
        return resp


class UserInfoResponse(Response):
    endpoint = 'userinfo'

    def __init__(self, conv, inut, sh, **kwargs):
        Response.__init__(self, conv, inut, sh, **kwargs)
        try:
            self.op_type = kwargs['op_type']
        except KeyError:
            self.op_type = ''
        else:
            del kwargs['op_type']
        self.msg_args = kwargs

    def construct_message(self):
        _kwargs = {'request': self.conv.events.last_item(EV_REQUEST)}
        try:
            _kwargs['authn'] = self.conv.events.last_item('HTTP_AUTHORIZATION')
        except NoSuchEvent:
            pass

        _kwargs.update(self.msg_args)
        _kwargs.update(self.op_args)

        resp = self.conv.entity.userinfo_endpoint(**_kwargs)
        save_response(self.conv.events, resp,
                      oic.message.OpenIDSchema,
                      oic.message.UserInfoErrorResponse)
        return resp


def factory(name):
    for fname, obj in inspect.getmembers(sys.modules[__name__]):
        if inspect.isclass(obj):
            if name == fname:
                return obj



from otest.aus.preproc import PMAP


def op_choice(base, flows):
    """
    Creates a list of test flows
    """
    _grp = "_"
    color = ['<button type="button" class="btn btn-default" title="Black"><span class="glyphicon glyphicon-play"></span></button>',  # INFORMATION
             '<button type="button" class="btn btn-success" title="Green"><span class="glyphicon glyphicon-ok-sign"></span></button>',  # OK
             '<button type="button" class="btn btn-warning" title="Yellow"><span class="glyphicon glyphicon-warning-sign"></span></button>',  # WARNING
             '<button type="button" class="btn btn-danger" title="Red"><span class="glyphicon glyphicon-minus-sign"></span></button>',  # ERROR
             '<button type="button" class="btn btn-danger" title="Red"><span class="glyphicon glyphicon-minus-sign"></span></button>',  # CRITICAL
             '<button type="button" class="btn btn-default" title="QuestionMark"><span class="glyphicon glyphicon-question-sign"></span></button>',  # INTERACTION
             '<button type="button" class="btn btn-default" title="QuestionMark"><span class="glyphicon glyphicon-question-sign"></span></button>',  # INCOMPLETE
             '<button type="button" class="btn btn-default" title="Grey"><span class="glyphicon glyphicon-record"></span></button>',  # NOT_APPLICABLE
             ]
    line = [
        '<table class="table table-hover table-bordered table-condensed">',
        '<tr><th>Status</th><th>Name</th><th>Description</th><th>Info</th></tr>']

    for grp, state, desc, tid in flows:
        b = tid.split("-", 2)[1]
        if not grp == _grp:
            _grp = grp
            _item = '<td colspan="4" class="text-center info"><b>{}</b></td>'.format(_grp)
            line.append(
                '<tr id="{}">{}</tr>'.format(b, _item))

        if state:
            _info = "<a href='{}test_info/{}' class=\"btn btn-info\" role=\"button\"><span class=\"glyphicon glyphicon-info-sign\"></span></a>".format(
                base, tid)
        else:
            _info = ''

        _stat = "<a href='{}{}'>{}</a>".format(base, tid, color[state])
        line.append(
            '<tr><td>{}</td><td style="vertical-align: middle;">({})</td><td style="vertical-align: middle;">{}</td><td>{}</td></tr>'.format(
                _stat, tid, desc, _info))
    line.append('</table>')

    return "\n".join(line)


ICONS = [
    ('<button type="button" class="btn btn-default" title="Black"><span class="glyphicon glyphicon-play"></span></button>',
     "The test has not been run yet."),
    ('<button type="button" class="btn btn-success" title="Green"><span class="glyphicon glyphicon-ok-sign"></span></button>',
     "The test succeeded."),
    ('<button type="button" class="btn btn-warning" title="Yellow"><span class="glyphicon glyphicon-warning-sign"></span></button>',
     "Warning, something was not as expected."),
    ('<button type="button" class="btn btn-danger"  title="Red"><span class="glyphicon glyphicon-minus-sign"></span></button>',
     "The test failed."),
    ('<button type="button" class="btn btn-default" title="QuestionMark"><span class="glyphicon glyphicon-question-sign"></span></button>',     
     "The test flow wasn't completed. This may have been expected or not."),
    ('<button type="button" class="btn btn-info"><span class="glyphicon glyphicon-info-sign"></span></button>',
     "Signals the fact that there is trace information available for the test."),
]


def legends():
    element = ["<table class=\"table table-condensed\" id=\"legends\">"]
    for icon, txt in ICONS:
        element.append("<tr><td>%s</td><td>%s</td></tr>" % (icon, txt))
    element.append('</table>')
    return "\n".join(element)


L2I = {"webfinger": 1, "discovery": 2, "registration": 3}
CM = {"n": "none", "s": "sign", "e": "encrypt"}


def display_profile(spec):
    el = ["<table class=\"table table-condensed\">"]
    p = spec.split('.')
    el.append("<tr><td>%s</td></tr>" % PMAP[p[0]])
    for mode in ["webfinger", "discovery", "registration"]:
        if p[L2I[mode]] == "T":
            el.append("<tr><td>Dynamic %s</td></tr>" % mode)
        else:
            el.append("<tr><td>Static %s</td></tr>" % mode)
    if len(p) > 6:
        if p[6] == 'T':
            el.append("<tr><td>Form post tests</td></tr>")
    if len(p) > 4:
        if p[4]:
            el.append("<tr><td>Crypto support %s</td></tr>" % [CM[x] for x in p[4]])
    if len(p) == 6:
        if p[5] == '+':
            el.append("<tr><td>Extra tests</td></tr>")
    el.append("</table>")

    return "\n".join(el)


def display_info(info):
    line = ['<table>']
    keys = list(info['tool'].keys())
    keys.sort()
    for key in keys:
        val = info['tool'][key]
        line.append('<tr><th>{}</th><td>{}</td></tr>'.format(key, val))
    line.append('</table>')
    return '\n'.join(line)



import copy
import inspect
import json
import logging
import time
import sys

import cherrypy
from jwkest import as_bytes
from oic.oauth2.message import ErrorResponse
from oic.oauth2.message import Message
from otest import Break, ConfigurationError

from otest.events import EV_EVENT
from otest.events import EV_EXCEPTION
from otest.events import EV_FUNCTION
from otest.events import EV_PROTOCOL_RESPONSE
from otest.events import EV_RESPONSE
from otest.verify import Verify

logger = logging.getLogger(__name__)


def print_result(resp):
    try:
        cl_name = resp.__class__.__name__
    except AttributeError:
        cl_name = ""
        txt = resp
    else:
        txt = json.dumps(resp.to_dict(), sort_keys=True, indent=2,
                         separators=(',', ': '))

    logger.info("{}: {}".format(cl_name, txt))


def request_with_client_http_session(instance, method, url, **kwargs):
    """Use the clients http session to make http request.
    Note: client.http_request function requires the parameters in reverse
    order (compared to the requests library): (method, url) vs (url, method)
    so we can't bind the instance method directly.
    """
    return instance.conv.entity.http_request(url, method)


def link(url):
    return '<a href="{}">link</a>'.format(url)


class Operation(object):
    _tests = {"pre": [], "post": []}
    message_cls = None

    def __init__(self, conv, inut, sh, test_id='', conf=None, funcs=None,
                 check_factory=None, cache=None, profile='', tool_conf=None,
                 **kwargs):
        self.conv = conv
        self.inut = inut
        self.sh = sh
        self.funcs = funcs or {}
        self.test_id = test_id
        self.conf = conf
        self.check_factory = check_factory
        self.cache = cache
        self.profile = profile
        self.tool_conf = tool_conf
        self.req_args = {}
        self.op_args = {}
        self.expect_exception = None
        self.expect_error = None
        self.sequence = []
        self.skip = False
        self.fail = False
        self.allowed_status_codes = [200]
        # detach
        self.tests = copy.deepcopy(self._tests)
        try:
            self.internal = kwargs['internal']
        except KeyError:
            self.internal = True
        self.unsupported = False

    def run(self, *args, **kwargs):
        return None

    def post_tests(self):
        if self.tests["post"]:
            cls_name = self.__class__.__name__
            _ver = Verify(self.check_factory, self.conv, cls_name=cls_name)
            _ver.test_sequence(self.tests["post"])

    def __call__(self, *args, **kwargs):
        if self.skip:
            return
        else:
            cls_name = self.__class__.__name__
            if self.tests["pre"] or self.tests["post"]:
                _ver = Verify(self.check_factory, self.conv, cls_name=cls_name)
            else:
                _ver = None

            if self.tests["pre"]:
                _ver.test_sequence(self.tests["pre"])

            res = self.run(*args, **kwargs)

            if res:
                return res

    def _setup(self):
        if self.skip:  # Don't bother
            return

        for op, arg in list(self.funcs.items()):
            try:
                op(self, arg)
            except ConfigurationError as err:
                _txt = "Configuration error: {}".format(err)
                self.conv.events.store(EV_EXCEPTION, _txt)
                raise cherrypy.HTTPError(message=_txt)
            except Exception as err:
                _txt = "Can't do {}".format(op)
                self.conv.events.store(EV_EXCEPTION, _txt)
                raise cherrypy.HTTPError(message=_txt)

                # self.conv.events.store(EV_OP_ARGS, self.op_args)

    def apply_profile(self, profile_map):
        try:
            kwargs = profile_map[self.__class__][self.profile[0]]
        except KeyError:
            return
        else:
            for op, arg in kwargs.items():
                op(self, arg)

    def op_setup(self):
        pass

    def setup(self, profile_map=None):
        """
        Order between setup methods are significant

        :param profile_map:
        :return:
        """
        if profile_map:
            self.apply_profile(profile_map)
        self.op_setup()
        self._setup()

    def catch_exception_and_error(self, func, **kwargs):
        """
        Seven possible use case
        1) The function returns an instance of the expected class
        2) The function returns a Message instance but not of the expected type
        3) The function returns an expected exception
        4) The function returns an unexpected exception
        5) The function returns an expected error message
        6) The function returns an unexpected error message
        7) Something else is returned.

        :param func:
        :param kwargs:
        :return:
        """
        res = None
        try:
            self.conv.events.store(EV_FUNCTION,
                                   {'name': func.__name__, 'kwargs': kwargs})
            res = func(**kwargs)
        except Exception as err:
            self.conv.events.store(
                EV_EXCEPTION, '{}:{}'.format(err.__class__.__name__, err))
            if not self.expect_exception:
                raise
            elif not err.__class__.__name__ == self.expect_exception:
                raise
            else:
                self.conv.events.store(
                    EV_EVENT,
                    'got expected exception {}'.format(self.expect_exception))
        else:
            if self.expect_exception:
                self.conv.events.store(
                    EV_EVENT, 'Expected exception did not occur')
                raise Exception(
                    "Expected exception '{}'.".format(self.expect_exception))

            if self.expect_error:
                #l = self.conv.events.events[-1].data
                try:
                    assert isinstance(res, ErrorResponse)
                except AssertionError:
                    logger.info(
                        'Expected error not received: {}'.format(res.to_dict()))
                    self.conv.events.store(
                        EV_EVENT, 'Expected error not received')
                else:
                    if res['error'] not in self.expect_error['error']:
                        self.conv.events.store(
                            EV_EVENT,
                            'Expected error not received: got {}'.format(
                                res['error']))
                        logger.info(
                            'Expected error not received: {}'.format(
                                res.to_dict()))
                    else:
                        self.conv.events.store(EV_EVENT, "Got expected error")
                        logger.info(
                            "Got expected error: {}".format(res.to_dict()))
                    try:
                        if self.expect_error['stop']:
                            raise Break('Stop')
                    except KeyError:
                        pass
            else:
                if isinstance(res, ErrorResponse):
                    logger.info(
                        'Unexpected error response: {}'.format(res.to_dict()))
                    self.conv.events.store(EV_EVENT, "Got unexpected error")
                    raise Break('Stop')

            if res:
                if isinstance(res, Message):
                    logger.info('Response: {}'.format(res.to_dict()))

                if isinstance(res, self.message_cls):
                    self.conv.events.store(EV_PROTOCOL_RESPONSE, res)
                elif isinstance(res, ErrorResponse):
                    pass
                elif isinstance(res, Message):
                    self.conv.events.store(EV_EVENT, "Got unexpected response")
                    raise Break('Unexpected response')
                else:
                    self.conv.events.store(EV_RESPONSE, res)
        return res

    def handle_response(self, *args):
        raise NotImplemented

    def handle_request(self, *args):
        raise NotImplemented


class Notice(Operation):
    pre_html = "message.html"

    def __init__(self, conv, inut, sh, **kwargs):
        Operation.__init__(self, conv, inut, sh, **kwargs)
        self.message = ""

    def args(self):
        return {'note': self.message, 'title': self.test_id}

    def __call__(self, *args, **kwargs):
        _msg = self.inut.pre_html[self.pre_html].format(**self.args())
        return as_bytes(_msg)


class Note(Notice):
    pre_html = "note.html"

    def op_setup(self):
        self.message = self.conv.flow["note"]

    def args(self):
        return {
            "next": link("{}continue?path={}&index={}".format(
                self.conv.entity.base_url, self.test_id, self.sh["index"])),
            "back": link(self.conv.entity.base_url),
            "note": self.message,
            "base": link(self.conv.entity.base_url),
            'header': "OpenID Certification OP Test",
            'title': self.test_id
        }


class TimeDelay(Operation):
    def __init__(self, conv, inut, sh, **kwargs):
        Operation.__init__(self, conv, inut, sh, **kwargs)
        self.delay = 30

    def __call__(self, *args, **kwargs):
        time.sleep(self.delay)
        return None


class ProtocolMessage(object):
    def __init__(self, conv, req_args, binding, msg_param=None):
        self.conv = conv
        self.entity = conv.entity
        self.req_args = req_args
        self.binding = binding
        self.msg_param = msg_param or {}
        self.response_args = {}
        self.op_args = {}

    def construct_message(self, *args):
        raise NotImplementedError()

    def handle_response(self, result, *args):
        raise NotImplementedError()


class VerifyConfiguration(Operation):
    def __init__(self, conv, inut, sh, **kwargs):
        Operation.__init__(self, conv, inut, sh, **kwargs)
        self.unsupported = ''

    def __call__(self, *args, **kwargs):
        if self.unsupported:
            raise Break(self.unsupported)


class SessionManagement(Operation):
    pre_html = "session_iframe.html"

    def __call__(self, *args, **kwargs):
        _msg = self.inut.pre_html[self.pre_html]
        # Need to replace client_id, session_state, issuer and
        # session_change_url
        client_id = self.conv.entity.client_id
        _msg = _msg.replace('{client_id}', client_id)
        session_state = ''
        _msg = _msg.replace('{session_state}', session_state)
        issuer = self.conv.entity.provider_info['issuer']
        _msg = _msg.replace('{issuer}', issuer)
        session_change_url = ''
        _msg = _msg.replace('{session_change_url}', session_change_url)
        return as_bytes(_msg)


def factory(name):
    for fname, obj in inspect.getmembers(sys.modules[__name__]):
        if inspect.isclass(obj):
            if name == fname:
                return obj



from oidctest.op import func
from oidctest.op import oper
from oidctest.op.client import Client
from oidctest.session import SessionHandler

from otest.aus.handling_ph import WebIh
from otest.conf_setup import OP_ORDER
from otest.conversation import Conversation
from otest.events import Events
from otest.flow import FlowState
from otest.prof_util import ProfileHandler

from oic.oic.message import factory
from oic.oic.message import ProviderConfigurationResponse
from oic.utils.authn.client import CLIENT_AUTHN_METHOD


def setup_conv():
    entity = Client(client_authn_method=CLIENT_AUTHN_METHOD,
                    verify_ssl=False)
    entity.provider_info = ProviderConfigurationResponse(
        authorization_endpoint="https://example.com",
    )

    cls_factories = {'': oper.factory}
    func_factory = func.factory

    flow_state = FlowState('flows', profile_handler=ProfileHandler,
                           cls_factories=cls_factories,
                           func_factory=func_factory,
                           display_order=OP_ORDER)
    iss = 'https://example.org'
    tag = 'foobar'
    session_handler = SessionHandler(iss, tag,
                                     flows=flow_state,
                                     tool_conf={})  # , rest=rest, **webenv)
    session_handler.iss = iss
    session_handler.tag = tag

    info = WebIh(session=session_handler, profile_handler=ProfileHandler)

    conv = Conversation([], entity, factory, callback_uris=[])
    conv.events = Events()
    conv.tool_config = {}
    return {'conv': conv, 'io': info}



import logging

from otest import exception_trace, Break
from otest.check import ERROR, State
from otest.check import WARNING
from otest.events import EV_CONDITION
from otest.result import Result
from otest.result import SIGN
from otest.summation import eval_state
from otest.summation import represent_result

__author__ = 'roland'

logger = logging.getLogger(__name__)


class InfoHandling(object):
    def __init__(self, flow_state, desc=None, profile_handler=None,
                 cache=None, session=None, **kwargs):
        self.flow_state = flow_state
        self.cache = cache
        self.profile_handler = profile_handler
        self.desc = desc
        self.session = session

    @property
    def profile(self):
        if self.session:
            return self.session.profile
        else:
            return ''

    def represent_result(self, events):
        return represent_result(events)

    def _store_error(self, where, err):
        if err:
            exception_trace(where, err, logger)

    def err_response(self, where, err):
        self._store_error(where, err)

    def get_err_type(self, test_id):
        errt = WARNING
        try:
            if self.session.profile.split('.')[0] in self.flow_state[
                    test_id]['MTI']:
                errt = ERROR
        except KeyError:
            pass
        return errt

    def log_fault(self, session, err, where, err_type=0):
        if err_type == 0:
            err_type = self.get_err_type(session)

        if "conv" in session:
            if err:
                session["conv"].events.store(EV_CONDITION,
                                             State("Fault", status=ERROR,
                                                   name=err_type,
                                                   message="{}".format(err)))
            else:
                session["conv"].events.store(
                    EV_CONDITION, State(
                        "Fault", status=ERROR,
                        name=err_type,
                        message="Error in %s" % where))


class ClIh(InfoHandling):
    def __init__(self, flows=None, profile='', desc='', profile_handler=None,
                 cache=None, session=None, **kwargs):
        InfoHandling.__init__(self, flows, profile=profile, desc=desc,
                              profile_handler=profile_handler,
                              cache=cache, session=session, **kwargs)

    def flow_list(self):
        pass

    def result(self):
        _state = eval_state(self.session["conv"].events)
        print(("{} {}".format(SIGN[_state], self.session["node"].name)))



import logging
import sys
import traceback

from otest import Break
from otest import ConditionError
from otest import exception_trace
from otest.check import CRITICAL
from otest.check import INTERACTION
from otest.events import EV_ASSERTION
from otest.events import EV_CONDITION
from otest.events import EV_FAULT

__author__ = 'roland'

logger = logging.getLogger(__name__)

LABELS = {
    'cid': 'WHERE', 'status': 'STATUS',
    'http_status': 'HTTP STATUS', 'message': 'INFO'
}


class MissingTest(Exception):
    pass


class Verify(object):
    def __init__(self, check_factory, conv, cls_name=''):
        self.check_factory = check_factory
        self.ignore_check = []
        self.exception = None
        self.conv = conv
        self.cls_name = cls_name

    def check_severity(self, stat):
        if stat.status in [CRITICAL, INTERACTION]:
            for attr, label in LABELS.items():
                try:
                    _val = getattr(stat, attr)
                except AttributeError:
                    pass
                else:
                    if _val:
                        self.conv.events.store(
                            EV_FAULT,
                            "{label}: {val}".format(val=_val, label=label))

            try:
                if not stat.mti:
                    raise Break(stat.message)
                else:
                    raise ConditionError(stat.message)
            except KeyError:
                pass

    def do_check(self, test, **kwargs):
        logger.debug("do_check({}, {})".format(test, kwargs))
        if isinstance(test, str):
            try:
                chk = self.check_factory(test)(**kwargs)
            except TypeError:
                raise MissingTest(test)
        else:
            chk = test(**kwargs)

        if chk.__class__.__name__ not in self.ignore_check:
            self.conv.events.store(EV_ASSERTION, chk.__class__.__name__)
            try:
                stat = chk(self.conv)
            except Exception as err:
                exception_trace('do_check', err, logger)
                raise
            else:
                if self.cls_name:
                    stat.context = self.cls_name

                self.conv.events.store(EV_CONDITION, stat,
                                       sender=self.__class__)
                self.check_severity(stat)

    def err_check(self, test, err=None, bryt=True):
        if err:
            self.exception = err
        chk = self.check_factory(test)()
        chk(self, self.conv.events.last_item(EV_CONDITION))
        if bryt:
            e = ConditionError("%s" % err)
            e.trace = "".join(traceback.format_exception(*sys.exc_info()))
            raise e

    def test_sequence(self, sequence):
        if isinstance(sequence, dict):
            for test, kwargs in list(sequence.items()):
                if not kwargs:
                    self.do_check(test)
                else:
                    self.do_check(test, **kwargs)
        else:
            for test in sequence:
                if isinstance(test, tuple):
                    test, kwargs = test
                    if not kwargs:
                        kwargs = {}
                else:
                    kwargs = {}
                self.do_check(test, **kwargs)
        return True



import inspect
import json
import sys

#from urllib.parse import urlencode
#from urllib.parse import urlparse
from future.backports.urllib.parse import urlparse
from future.backports.urllib.parse import urlencode

from otest import ConfigurationError
from otest.check import ERROR
from otest.check import get_id_tokens
from otest.check import State
from otest.events import EV_CONDITION
from otest.events import EV_RESPONSE
from otest.result import get_issuer
from otest.tool import get_redirect_uris

from oic.extension.message import make_software_statement
from oic.utils.keyio import KeyBundle

__author__ = 'roland'


class SetUpError(Exception):
    pass


def set_request_args(oper, args):
    oper.req_args.update(args)


def set_response_args(oper, args):
    oper.response_args.update(args)


def set_op_args(oper, args):
    oper.op_args.update(args)


def set_arg(oper, args):
    """

    :param oper: Operation instance
    :param args: dictionary with operation parameter as key and parameter
        value as value
    """
    for key, val in args.items():
        setattr(oper, key, val)


def expect_exception(oper, args):
    set_arg(oper, {'expect_exception': args})


def conditional_expect(oper, args):
    condition = args["condition"]

    res = True
    for key in list(condition.keys()):
        try:
            assert oper.req_args[key] in condition[key]
        except KeyError:
            pass
        except AssertionError:
            res = False

    for param in ['error', 'exception']:
        do_set = False
        try:
            if res == args["oper"]:
                do_set = True
        except KeyError:
            if res is True:
                do_set = True

        if do_set:
            try:
                setattr(oper, 'expect_{}'.format(param), args[param])
            except KeyError:
                pass


def set_expect_error(oper, args):
    set_arg(oper, {'expect_error': args})


def set_allowed_status_codes(oper, args):
    set_arg(oper, {'allowed_status_codes': args})


def set_time_delay(oper, args):
    set_arg(oper, {'delay': args})


def skip_operation(oper, arg):
    if oper.profile[0] in arg["flow_type"]:
        oper.skip = True


def conditional_expect_exception(oper, args):
    condition = args["condition"]
    exception = args["exception"]

    res = True
    for key in list(condition.keys()):
        try:
            assert oper.req_args[key] in condition[key]
        except KeyError:
            pass
        except AssertionError:
            res = False

    try:
        if res == args["oper"]:
            oper.expect_exception = exception
    except KeyError:
        if res is True:
            oper.expect_exception = exception


def add_post_condition(oper, args):
    for key, item in args.items():
        oper.tests['post'].append((key, item))


def add_pre_condition(oper, args):
    for key, item in args.items():
        oper.tests['pre'].append((key, item))


def clear_cookies(oper, args):
    oper.client.cookiejar.clear()


def set_uri(oper, args):
    ru = oper.conv.get_redirect_uris()[0]
    p = urlparse(ru)
    oper.req_args[args[0]] = "%s://%s/%s" % (p.scheme, p.netloc, args[1])


def get_base(base_url):
    """
    Make sure a '/' terminated URL is returned
    """
    part = urlparse(base_url)

    if part.path:
        if not part.path.endswith("/"):
            _path = part.path[:] + "/"
        else:
            _path = part.path[:]
    else:
        _path = "/"

    return "%s://%s%s" % (part.scheme, part.netloc, _path,)


def check_endpoint(oper, args):
    try:
        _ = oper.conv.entity.provider_info[args]
    except KeyError:
        oper.conv.events.store(
            EV_CONDITION,
            State("check_endpoint", status=ERROR,
                  message="{} not in provider configuration".format(args)))
        oper.skip = True


def cache_response(oper, arg):
    key = oper.conv.test_id
    oper.cache[key] = oper.conv.events.last_item(EV_RESPONSE)


def restore_response(oper, arg):
    key = oper.conv.test_id
    if oper.conv.events[EV_RESPONSE]:
        _lst = oper.cache[key][:]
        for x in oper.conv.events[EV_RESPONSE]:
            if x not in _lst:
                oper.conv.events.append(_lst)
    else:
        oper.conv.events.extend(oper.cache[key])

    del oper.cache[key]


def rm_claim_from_assertion(oper, arg):
    pass


def set_req_arg_token(oper, arg):
    oper.req_args["token_type_hint"] = arg
    oper.req_args['token'] = getattr(oper._token, arg)


def add_software_statement(oper, arg):
    argkeys = list(arg.keys())
    kwargs = {}

    tre = oper.conf.TRUSTED_REGISTRATION_ENTITY
    iss = tre['iss']
    kb = KeyBundle()
    kb.imp_jwks = json.load(open(tre['jwks']))
    kb.do_keys(kb.imp_jwks['keys'])
    oper.conv.entity.keyjar.add_kb(iss, kb)

    if arg['redirect_uris'] is None:
        kwargs['redirect_uris'] = oper.conv.entity.redirect_uris
    else:
        kwargs['redirect_uris'] = arg['redirect_uris']
    argkeys.remove('redirect_uris')

    if 'jwks_uri' in argkeys:
        if arg['jwks_uri'] is None:
            kwargs['jwks_uri'] = oper.conv.entity.jwks_uri
        else:
            kwargs['jwks_uri'] = arg['jwks_uri']
        argkeys.remove('jwks_uri')
    elif 'jwks' in argkeys:
        if arg['jwks'] is None:
            kwargs['jwks'] = {
                "keys": oper.conv.entity.keyjar.dump_issuer_keys("")}
        else:
            kwargs['jwks'] = arg['jwks']
        argkeys.remove('jwks')

    for a in argkeys:
        kwargs[a] = arg[a]

    oper.req_args['software_statement'] = make_software_statement(
        oper.conv.entity.keyjar, iss=iss, owner=iss, **kwargs)


def set_start_page(oper, args):
    _conf = oper.sh['test_conf']
    _url = _conf['start_page']
    _iss = oper.conv.entity.baseurl
    _params = _conf['params'].replace('<issuer>', _iss)
    _args = dict([p.split('=') for p in _params.split('&')])
    oper.start_page = _url + '?' + urlencode(_args)


def set_target(oper, args):
    oper.op_args['target'] = oper.conv.entity.provider_info['issuer']


def set_info_issuer(oper, args):
    oper.conv.info["issuer"] = oper.conv.get_tool_attribute("issuer")


def factory(name):
    for fname, obj in inspect.getmembers(sys.modules[__name__]):
        if inspect.isfunction(obj):
            if fname == name:
                return obj

    return None



#!/usr/bin/env python3
import importlib
import json
import logging
import os
import sys
import requests
import traceback

from os import listdir
from os.path import isdir
from os.path import isfile
from os.path import join

from future.backports.urllib.parse import parse_qs, urlparse

from mako.lookup import TemplateLookup

from oic import rndstr
from oic.utils.http_util import BadRequest
from oic.utils.http_util import extract_from_request
from oic.utils.http_util import get_or_post
from oic.utils.http_util import NotFound
from oic.utils.http_util import Response
from oic.utils.http_util import ServiceError
from oic.utils.http_util import SeeOther
from oic.utils import http_util

from otest.events import Events
from otest.events import EV_REQUEST
from otest.events import EV_RESPONSE
from otest.flow import FlowState
from otest.session import SessionHandler
from otest.rp.endpoints import static_mime
from otest.rp.handling import WebIh
from otest.rp.setup import as_arg_setup
from otest.rp.tool import WebTester

try:
    from requests.packages import urllib3
except ImportError:
    pass
else:
    urllib3.disable_warnings()

__author__ = 'roland'

ROOT = './'

LOOKUP = TemplateLookup(directories=[ROOT + 'htdocs'],
                        module_directory=ROOT + 'modules',
                        input_encoding='utf-8', output_encoding='utf-8')

logger = logging.getLogger("")


def setup_logging(logfile_name, log_level=logging.DEBUG):
    hdlr = logging.FileHandler(logfile_name)
    base_formatter = logging.Formatter(
        "%(asctime)s %(name)s:%(levelname)s %(message)s")

    hdlr.setFormatter(base_formatter)
    logger.addHandler(hdlr)
    logger.setLevel(log_level)


class JLog(object):
    def __init__(self, log, sid):
        self.logger = log
        self.id = sid

    def info(self, info):
        _dict = {'id': self.id}
        _dict.update(info)
        self.logger.info(json.dumps(_dict))

    def debug(self, info):
        _dict = {'id': self.id}
        _dict.update(info)
        self.logger.debug(json.dumps(_dict))

    def exception(self, info):
        _dict = {'id': self.id}
        _dict.update(info)
        self.logger.exception(json.dumps(_dict))

    def error(self, info):
        _dict = {'id': self.id}
        _dict.update(info)
        self.logger.error(json.dumps(_dict))

    def warning(self, info):
        _dict = {'id': self.id}
        _dict.update(info)
        self.logger.warning(json.dumps(_dict))


# def css(environ, event_db):
#     try:
#         info = open(environ["PATH_INFO"]).read()
#         resp = Response(info)
#     except (OSError, IOError):
#         resp = NotFound(environ["PATH_INFO"])
#
#     return resp


def start_page(environ, start_response, target):
    msg = open('start_page.html').read().format(target=target)
    resp = Response(msg)
    return resp(environ, start_response)


def make_entity(provider_cls, **kw_args):
    return provider_cls(**kw_args)


def absolute_url(url, startpage):
    if url.startswith('http'):
        return url

    (scheme, netloc, path, params, query, fragment) = urlparse(startpage)
    return '{}://{}{}'.format(scheme, netloc, url)


# =============================================================================


class Application(object):
    def __init__(self, base_url, **kwargs):
        self.base_url = base_url
        self.kwargs = kwargs
        self.events = Events()
        self.endpoints = {}
        self.session_conf = {}
        self.internal = kwargs['internal']

    def store_response(self, response):
        self.events.store(EV_RESPONSE, response.info())

    def wsgi_wrapper(self, environ, func, **kwargs):
        kwargs = extract_from_request(environ, kwargs)
        self.events.store(EV_REQUEST, kwargs)
        args = func(**kwargs)

        try:
            resp, state = args
            self.store_response(resp)
            return resp
        except TypeError:
            resp = args
            self.store_response(resp)
            return resp
        except Exception as err:
            logger.error("%s" % err)
            raise

    def handle(self, environ, tester, sid, path, qs=''):
        _sh = tester.sh
        if qs:
            msg = qs
        else:
            try:
                msg = get_or_post(environ)
            except AttributeError:
                msg = {}

        filename = self.kwargs['profile_handler'](_sh).log_path(
            sid=sid, test_id=_sh['conv'].test_id)

        _sh['conv'].entity_id = sid
        return tester.do_next(msg, filename,
                              profile_handler=self.kwargs['profile_handler'],
                              path=path)

    @staticmethod
    def pick_grp(name):
        return name.split('-')[1]

    @staticmethod
    def see_other_to_get(resp, sh):
        loc = resp.message
        res = sh['conv'].entity.server.http_request(loc, 'GET')
        return res

    def store_session_handler(self, sh):
        sid = rndstr(24)
        while sid in self.session_conf:
            sid = rndstr(24)
        sh['sid'] = sid
        self.session_conf[sid] = sh
        return sid

    def init_session(self, tester, sh, test_id=''):
        sid = self.store_session_handler(sh)
        # session['session_info'] = sh
        try:
            del self.session_conf[sid]['flow']
        except KeyError:
            pass

        try:
            args = sh['test_conf']
        except:
            args = {}

        args['test_id'] = test_id
        return tester.do_config(sid, **args)

    def run_test(self, tester, _path, _sid, environ, start_response):
        _op = '{} {}'.format(environ['REQUEST_METHOD'], _path)
        resp = tester.run(_path, sid=_sid, op=_op, **self.kwargs)
        if resp:
            logger.info(
                'Response class: {}'.format(resp.__class__.__name__))

        if isinstance(resp, requests.Response):
            try:
                loc = resp.headers['location']
            except KeyError:
                logger.info(
                    'Response type: {}, missing location'.format(
                        type(resp)))
                resp = ServiceError(
                    'Wrong response: {}:{}'.format(resp.status_code,
                                                   resp.text))
                return resp(environ, start_response), 0
            else:
                try:
                    tester.conv.events.store('Cookie',
                                             resp.headers['set-cookie'])
                except KeyError:
                    pass
                # For me !
                if loc.startswith(tester.base_url):
                    _path = loc[len(tester.base_url):]
                    if _path[0] == '/':
                        _path = _path[1:]
                    return 0, _path
                else:
                    if self.internal:
                        _url = absolute_url(loc,
                                            tester.sh['test_conf'][
                                                'start_page'])
                        logging.info('Redirect not to me => {}'.format(_url))
                        res = tester.conv.entity.server.http_request(_url)
                        logging.info('{} response'.format(res.status_code))
                        logging.debug('txt: {}'.format(res.text))
                        res = tester.display_test_list()
                        return res, 0
                    else:
                        res = SeeOther(loc)
                        return res(environ, start_response), 0
        elif resp is True or resp is False or resp is None:
            return tester.display_test_list(), 0
        else:
            return resp(environ, start_response), 0

    # publishes the OP endpoints
    def application(self, environ, start_response):
        session = environ['beaker.session']

        jlog = JLog(logger, session.id)
        path = environ.get('PATH_INFO', '').lstrip('/')
        jlog.info({"remote_addr": environ["REMOTE_ADDR"],
                   "path": path})

        # self.events.store(EV_REQUEST, path)

        try:
            sh = session['session_info']
        except KeyError:
            sh = SessionHandler(**self.kwargs)
            #sh.session_init()
            session['session_info'] = sh

        info = WebIh(session=sh, **self.kwargs)
        info.environ = environ
        info.start_response = start_response

        tester = WebTester(info, sh, **self.kwargs)

        if 'path' in self.kwargs and path.startswith(self.kwargs['path']):
            _path = path[len(kwargs['path']) + 1:]
        else:
            _path = path

        if _path == "robots.txt":
            return static_mime("static/robots.txt", environ, start_response)
        elif _path.startswith("static/"):
            return static_mime(_path, environ, start_response)

        if _path == "list":
            try:
                qs = parse_qs(get_or_post(environ))
            except Exception as err:
                jlog.error({'message': err})
                qs = {}
            else:
                if qs:
                    sh['test_conf'] = dict([(k, v[0]) for k, v in qs.items()])
                    # self.session_conf[sh['sid']] = sh
                else:
                    return self.init_session(tester, sh)

                logger.info('test_conf: {}'.format(sh['test_conf']))

                if 'start_page' not in sh['test_conf']:
                    resp = BadRequest('You MUST provide a start_page')
                    return resp(environ, start_response)

            info.profile = tester.sh.profile = qs['response_type'][0]
            sh.session_init()

            if 'test_id' in qs:
                (res, _path) = self.run_test(tester, qs['test_id'][0],
                                             sh['sid'], environ,
                                             start_response)
                if res:
                    return res
            else:
                res = tester.display_test_list()
                return res
        elif _path == '' or _path == 'config':
            return self.init_session(tester, sh)
        elif _path in self.kwargs['flows'].keys():  # Run flow
            # Will use the same test configuration
            try:
                _ = tester.sh['test_conf']
            except KeyError:
                return self.init_session(tester, sh)

            try:
                _sid = tester.sh['sid']
            except KeyError:
                _sid = self.store_session_handler(sh)

            # First time around this should not be set
            try:
                _ = self.session_conf[_sid]['flow']
            except KeyError:
                pass
            else:
                return self.init_session(tester, sh, _path)

            (res, _path) = self.run_test(tester, _path, _sid, environ,
                                         start_response)
            if res:
                return res
        elif _path == 'display':
            return info.flow_list()
        elif _path == "opresult":
            try:
                _display_path = '/{}/display'.format(self.kwargs['path'])
            except KeyError:
                _display_path = '/display'
            resp = SeeOther(
                "{}#{}".format(_display_path,
                               self.pick_grp(sh['conv'].test_id)))
            return resp(environ, start_response)
        elif _path.startswith("test_info"):
            p = _path.split("/")
            try:
                return info.test_info(p[1])
            except KeyError:
                return info.not_found()
        elif _path == 'all':
            for test_id in sh['flow_names']:
                resp = tester.run(test_id, **self.kwargs)
                if resp is True or resp is False:
                    continue
                elif resp:
                    return resp(environ, start_response)
                else:
                    resp = ServiceError('Unkown service error')
                    return resp(environ, start_response)
            return tester.display_test_list()

        # Whatever gets here should be of the form <session_id>/<path>
        try:
            sid, _path = _path.split('/', 1)
        except ValueError:
            pass
        else:
            if _path.startswith("static/"):
                return static_mime(_path, environ, start_response)

            try:
                _sh = self.session_conf[sid]
            except KeyError:
                resp = ServiceError("Unknown session")
                return resp(environ, start_response)

            tester.sh = _sh
            if 'HTTP_AUTHORIZATION' in environ:
                _sh['conv'].events.store('HTTP_AUTHORIZATION',
                                         environ['HTTP_AUTHORIZATION'])
            _p = _path.split('?')
            _sh['conv'].events.store('http request', '{} /{}'.format(
                environ['REQUEST_METHOD'], _path))
            if _p[0] in _sh['conv'].entity.endpoints():
                resp = self.handle(environ, tester, sid, *_p)
                self.session_conf[sid] = tester.sh
                # The only redirect should be the one to the redirect_uri
                if isinstance(resp, SeeOther):
                    if self.internal:
                        # res = self.see_other_to_get(resp, sh)
                        # res is probably a redirect
                        # send the user back to the test list page
                        return info.flow_list()
                    else:
                        return resp(environ, start_response)
                elif isinstance(resp, Response):
                    return resp(environ, start_response)
                else:
                    return resp

            for endpoint, service in self.endpoints.items():
                if _path == endpoint:
                    jlog.info({"service": service})
                    try:
                        resp = self.handle(environ, tester, sid, service)
                        return resp(environ, start_response)
                    except Exception as err:
                        print("%s" % err)
                        message = traceback.format_exception(*sys.exc_info())
                        print(message)
                        jlog.exception(err)
                        resp = ServiceError("%s" % err)
                        return resp(environ)

        jlog.debug({"unknown side": path})
        resp = NotFound("Couldn't find the side you asked for!")
        return resp(environ, start_response)


def key_handling(key_dir):
    if isdir(key_dir):
        only_files = [f for f in listdir(key_dir) if isfile(join(key_dir, f))]
    else:
        os.makedirs(key_dir)
        only_files = []

    if not only_files:
        only_files = ['one.pem']
        for fil in only_files:
            key = RSA.generate(2048)
            f = open(join(key_dir, fil), 'w')
            f.write(key.exportKey('PEM').decode('utf8'))
            f.close()

    return {key_dir: only_files}


# def find_allowed_algorithms(metadata_file, ic):
#     mds = MetadataStore(ic.attribute_converters, ic,
#                         disable_ssl_certificate_validation=True)
#
#     mds.imp([{
#         "class": "saml2.mdstore.MetaDataFile",
#         "metadata": [(metadata_file,)]}])
#
#     md = mds.metadata[metadata_file]
#     ed = list(md.entity.values())[0]
#     res = {"digest_algorithms":[], "signing_algorithms":[]}
#
#     for elem in ed['extensions']['extension_elements']:
#         if elem['__class__'] == '{}&DigestMethod'.format(
# algsupport.NAMESPACE):
#             res['digest_algorithms'].append(elem['algorithm'])
#         elif elem['__class__'] == '{}&SigningMethod'.format(
#                 algsupport.NAMESPACE):
#             res['signing_algorithms'].append(elem['algorithm'])
#
#     return res


if __name__ == '__main__':
    import argparse
    from beaker.middleware import SessionMiddleware
    from Cryptodome.PublicKey import RSA

    from cherrypy import wsgiserver
    from cherrypy.wsgiserver.ssl_builtin import BuiltinSSLAdapter

    parser = argparse.ArgumentParser()
    parser.add_argument('-d', dest='debug', action='store_true')
    parser.add_argument(
        '-k', dest='insecure', action='store_true',
        help='whether or not TLS certificate verification should be performed')
    parser.add_argument(
        '-H', dest='hostname',
        help='If running behind a proxy this is the external name of the host')
    parser.add_argument('-s', dest='tls', action='store_true',
                        help='Whether the server should handle SSL/TLS')
    parser.add_argument(
        '-f', dest='flowdir',
        help='Directory where test descriptions in JSON format can be found')
    parser.add_argument('-r', dest='rsa_key_dir', default='keys')
    parser.add_argument('-l', dest='logfile')
    parser.add_argument(
        '-i', dest='internal', action='store_true',
        help='Whether the server should handle all communication internally')
    parser.add_argument('-m', dest='path2port')
    parser.add_argument('-w', dest='cwd', help='change working directory')
    parser.add_argument(
        '-P', dest='port', help='Which port the test instance should listen at')
    parser.add_argument('-O', dest='op_profiles',
                        help='Possible OP (=test tool) profiles')
    parser.add_argument(
        '-c', dest="ca_certs",
        help="CA certs to use to verify HTTPS server certificates, "
             "if HTTPS is used and no server CA certs are defined then "
             "no cert verification will be done")
    parser.add_argument(
        '-x', dest='xport', help='ONLY for testing')
    parser.add_argument(dest="config")
    args = parser.parse_args()

    if args.logfile:
        setup_logging(args.logfile)
    else:
        setup_logging('tt.log')

    session_opts = {
        'session.type': 'memory',
        'session.cookie_expires': True,
        'session.auto': True,
        # 'session.key': "{}.beaker.session.id".format(
        #     urlparse(_base).netloc.replace(":", "."))
    }

    sys.path.insert(0, ".")
    config = importlib.import_module(args.config)
    tool_args = config.TOOL_ARGS

    flows = FlowState(args.flowdir, tool_args['profile_handler'],
                      tool_args['cls_factories'], tool_args['func_factory'],
                      display_order=['Code', 'Token'])

    # Create necessary keys if I don't already have them
    keys = key_handling('keys')

    if args.insecure:
        disable_validation = True
    else:
        disable_validation = False

    if args.cwd:
        base_dir = args.cwd
    else:
        base_dir = os.getcwd()

    as_args, key_args = as_arg_setup(args, lookup=LOOKUP, config=config)

    _op_profiles = json.load(open(args.op_profiles))

    kwargs = {"base_url": as_args['name'], 'flows': flows,
              'order': ["Code", "Token"],
              "msg_factory": tool_args['cls_factories'],
              "check_factory": tool_args['chk_factory'], 'conf': config,
              "cache": {}, 'op_profiles': _op_profiles,
              "profile_handler": tool_args['profile_handler'], 'map_prof': None,
              'lookup': LOOKUP, 'make_entity': make_entity,
              'base_dir': base_dir, 'signing_key': keys,
              'provider_cls': tool_args['provider'], 'as_args': as_args,
              'response_cls': http_util.Response, 'internal': args.internal
              }

    if args.ca_certs:
        kwargs['ca_certs'] = args.ca_certs

    if args.path2port:
        kwargs['path'] = as_args['instance_path']

    _app = Application(base=as_args['name'], **kwargs)
    _app.endpoints = {
        '.well-known/openid-configuration': 'providerinfo_endpoint'
    }

    if args.xport:
        _port = int(args.xport)
    else:
        _port = int(as_args['instance_port'])

    # Initiate the web server
    SRV = wsgiserver.CherryPyWSGIServer(
        ('0.0.0.0', _port),
        SessionMiddleware(_app.application, session_opts))

    if args.tls:
        from cherrypy.wsgiserver.ssl_builtin import BuiltinSSLAdapter

        SRV.ssl_adapter = BuiltinSSLAdapter(config.SERVER_CERT,
                                            config.SERVER_KEY,
                                            config.CERT_CHAIN)
        extra = "using SSL/TLS"
    else:
        extra = ""

    print('issuer: {}'.format(as_args['name']))
    txt = "RP test tool started {}.".format(extra)
    logger.info(txt)
    print(txt)

    try:
        SRV.start()
    except KeyboardInterrupt:
        SRV.stop()



import json
import os
import re

import logging
from six import text_type

from otest import Done
from otest import Unknown
from otest.func import factory as ofactory
from otest.summation import completed
from otest.summation import eval_state
from otest.summation import represent_result

from otest.prof_util import from_profile

logger = logging.getLogger(__name__)

PAT = re.compile('\${([A-Z_0-9]*)}')

ABBR = {
    "code": 'C',
    "id_token": 'I',
    "id_token token": 'IT',
    "code id_token": 'CI',
    "code token": 'CT',
    "code id_token token": 'CIT',
    "dynamic": 'DYN',
    "configuration": 'CNF'
}

EXP = dict([(v, k) for k, v in ABBR.items()])

GRPS = [
    "Discovery", "Dynamic Client Registration",
    "Response Type and Response Mode", "claims Request Parameter",
    "request_uri Request Parameter", "scope Request Parameter",
    "nonce Request Parameter", 'redirect_uri Request Parameter',
    'prompt Request Parameter', 'request Request Parameter',
    "Client Authentication", "OAuth behaviors",
    "ID Token", "Key Rotation", "Claims Types", "UserInfo Endpoint",
    'Misc Request Parameters', 'Key Rotation', 'Access Token',
    "Response Type", 'Response Mode'
]


def replace_with_url(txt, links):
    for m in PAT.findall(txt):
        try:
            _url = links['URL'][m]
        except KeyError:
            pass
        else:
            txt = txt.replace('${%s}' % m, _url)

    return txt


def replace_with_link(txt, links):
    for m in PAT.findall(txt):
        try:
            _url, tag = links['LINK'][m]
        except KeyError:
            pass
        else:
            _li = replace_with_url(_url, links)
            _href = '<a href="{}">{}</a>'.format(_li, tag)
            txt = txt.replace('${%s}' % m, _href)
    return txt


class Flow(object):
    def __init__(self, fdir, profile_handler):
        self.fdir = fdir
        self.profile_handler = profile_handler

    def __getitem__(self, tid):
        """
        Get the flow description given a test ID

        :param tid: The test ID
        :return: A dictionary representation of the description
        """

        fname = os.path.join(self.fdir, tid + '.json')
        fp = open(fname, 'r')
        try:
            _info = json.load(fp)
        except Exception as err:
            logger.error(err)
            raise KeyError(tid)
        finally:
            fp.close()

        return _info

    def items(self):
        """
        Return all flow descriptions.
        It is assumed that all files with names that has the postfix '.json'
        prepresents flow descriptions.

        :return:
        """
        for fn in os.listdir(self.fdir):
            if fn.endswith('.json'):
                sfn = fn[:-5]
                yield ((sfn, self[sfn]))

    def keys(self):
        """
        Return all Test IDs
        :return: list of test IDs
        """
        for fn in os.listdir(self.fdir):
            if fn.endswith('.json'):
                yield (fn[:-5])

    def pick(self, key, value):
        """
        Pick a number of test descriptions base on a key,value pair.

        :param key:
        :param value:
        :return:
        """
        tids = []
        for tid, spec in self.items():
            try:
                _val = spec[key]
            except KeyError:
                pass
            else:
                if value == _val:
                    tids.append(tid)
        return tids

    def matches_profile(self, profile):
        """
        Return a list of test IDs that all match the profile
        :param profile:
        :return:
        """

        _tids = []
        _use = from_profile(profile)
        for tid, spec in self.items():
            if match_usage(spec, **_use):
                _tids.append(tid)
        return _tids

    def mandatory_to_implement(self, tid, profile):
        _use = from_profile(profile)
        _use['return_type'] = _use['return_type'][0]
        spec = self[tid]
        try:
            _mti = spec["MTI"]
        except KeyError:
            pass
        else:
            if _use['return_type'][0] in _mti:
                if _use['register'] and 'DYN' in _mti:
                    if _use['discover'] and 'CNF' in _mti:
                        return True
        return False

    def _profile_info(self, test_id, session):
        if self.profile_handler:
            ph = self.profile_handler(session)
            try:
                return ph.get_profile_info(test_id)
            except Exception as err:
                raise
        return {}

    def __contains__(self, item):
        fname = os.path.join(self.fdir, item + '.json')
        return os.path.isfile(fname)


# ==============================================================================


def _get_cls(name, factories, use=''):
    if use:
        try:
            cls = factories[use](name)
        except Unknown:
            pass
        else:
            return cls

    try:
        cls = factories[''](name)
    except Unknown:
        raise Exception("Unknown Class: '{}'".format(name))

    return cls


def _get_func(dic, func_factory):
    """
    Convert function names into function references

    :param dic: A key, value dictionary where keys are function names
    :param func_factory: Factory function used to find functions
    :return: A dictionary with the keys replace with references to functions
    """
    res = {}
    for fname, val in dic.items():
        func = func_factory(fname)
        if func is None:
            func = ofactory(fname)

        if func is None:
            raise Exception("Unknown function: '{}'".format(fname))
        res[func] = val

    return res


class RPFlow(Flow):
    def __init__(self, fdir, profile_handler, cls_factories, func_factory,
                 use=''):
        Flow.__init__(self, fdir, profile_handler)
        self.cls_factories = cls_factories
        self.func_factory = func_factory
        self.use = use

    def expanded_conf(self, tid):
        """

        :param test_id:
        :return:
        """
        spec = self[tid]
        seq = []
        for oper in spec["sequence"]:
            if isinstance(oper, dict):  # Must be only one key, value item
                if len(oper) > 1:
                    raise SyntaxError(tid)
                key, val = list(oper.items())[0]
                try:
                    seq.append((_get_cls(key, self.cls_factories, self.use),
                                _get_func(val, self.func_factory)))
                except Exception:
                    print('tid:{}'.format(tid))
                    raise
            else:
                try:
                    seq.append(_get_cls(oper, self.cls_factories, self.use))
                except Exception:
                    print('tid:{}'.format(tid))
                    raise
        seq.append(Done)
        spec["sequence"] = seq

        return spec


def match_usage(spec, **kwargs):
    try:
        _usage = spec['usage']
    except KeyError:
        return True
    else:
        for key, allowed in _usage.items():
            try:
                val = kwargs[key]
            except KeyError:
                return False
            else:
                if key == 'return_type':
                    # val can be list of one string or just a string
                    if isinstance(val, text_type):
                        if val not in allowed:
                            return False
                    elif val[0] not in allowed:
                        return False
                else:
                    if isinstance(allowed, bool):
                        if val is not allowed:
                            return False
                    else:
                        if val not in allowed:
                            return False
    return True


def get_return_type(prof):
    try:
        return prof.split('.')[0]
    except AttributeError:
        return prof[0]


def get_category(usage):
    if 'extra' in usage:
        return '[Extra]'
    if 'register' in usage:
        return '[Dynamic]'
    if 'discover' in usage:
        return '[Config]'

    if 'return_type' not in usage:
        return '[Basic, Implicit, Hybrid]'
    else:
        li = []
        _rt = usage['return_type']
        if 'C' in _rt:
            li.append('Basic')
        if 'IT' in _rt or 'I' in _rt:
            li.append('Implicit')
        if 'CT' in _rt or 'CI' in _rt or 'CIT' in _rt:
            li.append('Hybrid')
        return '[' + ','.join(li) + ']'


class FlowState(RPFlow):
    def __init__(self, fdir, profile_handler, cls_factories, func_factory,
                 display_order, use=''):
        RPFlow.__init__(self, fdir, profile_handler, cls_factories,
                        func_factory, use=use)
        self.test_info = {}
        self.display_order = display_order
        self.complete = {}

    def store_test_info(self, tester):
        _conv = tester.conv
        tinfo = self._test_info(_conv.test_id, _conv.events, _conv.index,
                                tester.sh)
        self.test_info[_conv.test_id] = tinfo
        return tinfo

    def _test_info(self, test_id, events, index, session):
        _info = {
            'test_id': test_id,
            "events": events,
            "index": index,
            "test_output": events.get('condition'),
            "state": eval_state(events),
            "complete": completed(events),
            "result": represent_result(events)
        }

        try:
            _info["descr"] = self[test_id]["desc"]
        except KeyError:
            _info['descr'] = session['flow']['desc']

        if _info['complete']:
            self.complete[test_id] = True

        _info['profile_info'] = self._profile_info(test_id, session)

        return _info

    def display_info(self, tids):
        """
        Return information to be used in UIX display
        :param tids: List of TestIDs
        :return:
        """

        interim = {}
        for tid in tids:
            _spec = self[tid]
            try:
                _state = self.test_info[tid]['state']
            except KeyError:
                _state = 0

            try:
                interim[_spec['group']].append((_state, _spec['desc'], tid))
            except KeyError:
                interim[_spec['group']] = [(_state, _spec['desc'], tid)]

        res = []
        for grp in self.display_order:
            try:
                _s = sorted(interim[grp], key=lambda x: x[2])
            except KeyError:
                continue
            else:
                for x in _s:
                    y = [grp]
                    y.extend(x)
                    res.append(y)
        return res



from otest import summation
from otest.events import row


def do_assertions(out):
    return summation.condition(out, True)


def trace_output(events):
    """

    """
    element = ['<table class="table table-bordered table-condensed">']
    start = 0
    for event in events:
        if not start:
            start = event.timestamp
        # element.append(layout(start, event))
        element.append(row(start, event))
    element.append("</table>")
    return "\n".join(element)


def profile_output(pinfo, version=''):
    element = ['<table class="table table-condensed">']
    for key, val in pinfo.items():
        element.append("<tr><th>%s</th><td>%s</td></tr>" % (key, val))
    element.append('</table>')
    return "\n".join(element)



import logging
import os

from oic.utils.http_util import BadRequest
from oic.utils.http_util import SeeOther

from otest.events import EV_HTTP_ARGS
from otest.result import safe_url

__author__ = 'roland'

logger = logging.getLogger(__name__)


class WebApplication(object):
    def __init__(self, sessionhandler, webio, webtester, check, webenv,
                 pick_grp, path=''):
        self.sessionhandler = sessionhandler
        self.webio = webio
        self.webtester = webtester
        self.check = check
        self.webenv = webenv
        self.pick_grp = pick_grp
        self.path = path

    def application(self, environ, start_response):
        logger.info("Connection from: %s" % environ["REMOTE_ADDR"])
        session = environ['beaker.session']

        path = environ.get('PATH_INFO', '').lstrip('/')
        logger.info("path: %s" % path)

        try:
            sh = session['session_info']
        except KeyError:
            sh = self.sessionhandler(**self.webenv)
            sh.session_init()
            session['session_info'] = sh

        info = self.webio(session=sh, **self.webenv)
        info.environ = environ
        info.start_response = start_response
        tester = self.webtester(info, sh, **self.webenv)
        tester.check_factory = self.check.factory

        if path == "robots.txt":
            return info.static("static/robots.txt")
        elif path == "favicon.ico":
            return info.static("static/favicon.ico")
        elif path.startswith("static/"):
            return info.static(path)
        elif path.startswith("jwks/"):
            return info.static(path)
        elif path.startswith("export/"):
            return info.static(path)

        if self.path and path.startswith(self.path):
            _path = path[len(self.path)+1:]
        else:
            _path = path

        if _path == "":  # list
            return tester.display_test_list()

        if _path == "logs":
            return info.display_log("log", issuer="", profile="", testid="")
        elif _path.startswith("log"):
            if _path == "log" or _path == "log/":
                try:
                    _iss = self.webenv['client_info']["provider_info"]["issuer"]
                except KeyError:
                    _iss = self.webenv['tool_conf']['issuer']

                parts = [safe_url(_iss)]
            else:
                parts = []
                while _path != "log":
                    head, tail = os.path.split(_path)
                    # tail = tail.replace(":", "%3A")
                    # if tail.endswith("%2F"):
                    #     tail = tail[:-3]
                    parts.insert(0, tail)
                    _path = head

            return info.display_log("log", *parts)
        elif _path.startswith("tar"):
            _path = _path.replace(":", "%3A")
            return info.static(_path)

        if _path == "reset":
            sh.reset_session()
            return info.flow_list()
        elif _path == "pedit":
            try:
                return info.profile_edit()
            except Exception as err:
                return info.err_response("pedit", err)
        elif _path == "profile":
            return tester.set_profile(environ)
        elif _path.startswith("test_info"):
            p = _path.split("/")
            try:
                return info.test_info(p[1])
            except KeyError:
                return info.not_found()
        elif _path == "continue":
            resp = tester.cont(environ, self.webenv)
            session['session_info'] = info.session
            if resp:
                return resp
            else:
                resp = SeeOther(
                    "{}display#{}".format(self.webenv['base_url'],
                                           self.pick_grp(sh['conv'].test_id)))
                return resp(environ, start_response)
        elif _path == 'display':
            return info.flow_list()
        elif _path == "opresult":
            resp = SeeOther(
                "{}display#{}".format(self.webenv['base_url'],
                                     self.pick_grp(sh['conv'].test_id)))
            return resp(environ, start_response)
        # expected _path format: /<testid>[/<endpoint>]
        elif _path in sh["tests"]:
            resp = tester.run(_path, **self.webenv)
            session['session_info'] = info.session

            if resp is False or resp is True:
                pass
            elif isinstance(resp, list):
                return resp

            try:
                #  return info.flow_list()
                resp = SeeOther(
                    "{}display#{}".format(
                        self.webenv['client_info']['base_url'],
                        self.pick_grp(sh['conv'].test_id)))
                return resp(environ, start_response)
            except Exception as err:
                logger.error(err)
                raise
        elif _path in ["authz_cb", "authz_post"]:
            if _path == "authz_cb":
                _conv = sh["conv"]
                try:
                    response_mode = _conv.req.req_args["response_mode"]
                except KeyError:
                    response_mode = ""

                # Check if fragment encoded
                if response_mode == "form_post":
                    pass
                else:
                    try:
                        response_type = _conv.req.req_args["response_type"]
                    except KeyError:
                        response_type = [""]

                    if response_type == [""]:  # expect anything
                        if environ["QUERY_STRING"]:
                            pass
                        else:
                            return info.opresult_fragment()
                    elif response_type != ["code"]:
                        # but what if it's all returned as a query anyway ?
                        try:
                            qs = environ["QUERY_STRING"]
                        except KeyError:
                            pass
                        else:
                            _conv.events.store(EV_HTTP_ARGS, qs)
                            _conv.query_component = qs

                        return info.opresult_fragment()

            try:
                resp = tester.async_response(self.webenv["conf"])
            except Exception as err:
                return info.err_response("authz_cb", err)
            else:
                if resp is False or resp is True:
                    pass
                elif not isinstance(resp, int):
                    return resp

                try:
                    # return info.flow_list()
                    resp = SeeOther(
                        "{}display#{}".format(
                            self.webenv['client_info']['base_url'],
                            self.pick_grp(sh['conv'].test_id)))
                    return resp(environ, start_response)
                except Exception as err:
                    logger.error(err)
                    raise
        else:
            resp = BadRequest()
            return resp(environ, start_response)



import json

from otest.check import OK
from otest.check import WARNING
from otest.check import ERROR
from otest.check import CRITICAL
from otest.events import EV_CONDITION, EV_HTTP_INFO
from otest.events import EV_HTTP_RESPONSE
from otest.events import EV_PROTOCOL_RESPONSE
from otest.events import EV_FAULT
from otest.events import EV_HANDLER_RESPONSE
from otest.events import EV_HTTP_ARGS
from otest.events import EV_HTTP_RESPONSE_HEADER
from otest.events import EV_OPERATION
from otest.events import EV_PROTOCOL_REQUEST
from otest.events import EV_REDIRECT_URL
from otest.events import EV_REPLY
from otest.events import EV_REQUEST
from otest.events import EV_REQUEST_ARGS
from otest.events import EV_RESPONSE
from otest.events import EV_RESPONSE_ARGS
from otest.events import EV_SEND
from otest.events import EV_URL

__author__ = 'roland'


def adjust_lines(lines, maxlen=120):
    res = []
    for line in lines.split('\n'):
        if len(line) <= maxlen:
            res.append(line)
        else:
            n = maxlen
            for l in [line[i:i + n] for i in range(0, len(line), n)]:
                res.append(l)
    return res


def print_message(message_as_dict):
    return adjust_lines(json.dumps(message_as_dict, sort_keys=True, indent=2,
                                   separators=(',', ': ')))


BG_COLOR = {
    OK: "background-color:lightgreen;",
    WARNING: "background-color:lightyellow;",
    ERROR: "background-color:lightcoral;",
    CRITICAL: "background-color:lightcoral;"
}


def do_condition(event):
    return ['<li style="{}">{}</li>'.format(BG_COLOR[event.data.status],
                                            event.data.test_id)]


def do_protocol_response(event):
    res = ['<h4>Response</h4>', '<pre>']
    res.extend(print_message(event.data.to_dict()))
    res.append('</pre>')
    return res


def do_http_info(event):
    d = event.data
    res = ['<h4>Response</h4>',
           '<table border="1" style="width:100%">',
           '<tr><td>status</td><td>{}</td></tr>'.format(d['status']),
           '<tr><td>headers</td><td><pre>[']

    for x in d['headers']:
        res.extend(adjust_lines('{}'.format(x)))

    res.append(']</pre></td></tr>')
    res.append('<tr><td>message</td><td><pre>')
    try:
        res.extend(print_message(json.loads(d['message'])))
    except ValueError:
        res.extend(adjust_lines(d['message']))
    res.append('</pre></td></tr>')
    res.extend(['</table>'])
    return res


def do_http_response(event):
    d = event.data
    res = ['<h4>Response</h4>',
           '<table border="1" style="width:100%">',
           '<tr><td>status</td><td>{}</td></tr>'.format(d.status_code),
           '<tr><td>headers</td><td><pre>[']

    for x in d.headers:
        res.extend(adjust_lines('{}'.format(x)))

    res.append(']</pre></td></tr>')
    res.append('<tr><td>message</td><td><pre>')
    try:
        res.extend(print_message(json.loads(d.text)))
    except ValueError:
        res.extend(adjust_lines(d.text))
    res.append('</pre></td></tr>')
    res.extend(['</table>'])
    return res


def do_protocol_request(event):
    res = ['<h4>Request</h4>', '<pre>']
    res.extend(print_message(event.data.to_dict()))
    res.append('</pre>')
    return res


def do_request(event):
    res = ['<h4>Request</h4>', '<pre>']
    try:
        res.extend(print_message(json.loads(event.data)))
    except ValueError:
        res.extend(adjust_lines(event.data))
    res.append('</pre>')
    return res


def do_fault(event):
    return ()


def do_handler_response(event):
    return ()


def do_http_args(event):
    return []


def do_http_response_header(event):
    return []


def do_operation(event):
    return []


def do_request_args(event):
    return []


def do_redirect_url(event):
    return []


def do_reply(event):
    return []


def do_response(event):
    return []


def do_response_args(event):
    return []


def do_send(event):
    return []


def do_url(event):
    return []


def do_path(event):
    return ['<hr>',
            '<h3>ENDPOINT: {}</h3>'.format(event.data),
            '<hr>']


FUNC_MAP = {
    EV_CONDITION: do_condition,
    EV_FAULT: do_fault,
    EV_HANDLER_RESPONSE: do_handler_response,
    EV_HTTP_ARGS: do_http_args,
    EV_HTTP_INFO: do_http_info,
    EV_HTTP_RESPONSE: do_http_response,
    EV_HTTP_RESPONSE_HEADER: do_http_response_header,
    EV_OPERATION: do_operation,
    EV_PROTOCOL_RESPONSE: do_protocol_response,
    EV_PROTOCOL_REQUEST: do_protocol_request,
    EV_REDIRECT_URL: do_redirect_url,
    EV_REPLY: do_reply,
    EV_REQUEST: do_request,
    EV_REQUEST_ARGS: do_request_args,
    EV_RESPONSE: do_response,
    EV_RESPONSE_ARGS: do_response_args,
    EV_SEND: do_send,
    EV_URL: do_url,
    'path': do_path
}


def display(events):
    last = ''
    lines = []
    request = None
    for event in events:
        if event.typ != last:
            if event.typ == EV_CONDITION:
                lines.append('<ul>')
            elif last == EV_CONDITION:
                lines.append('</ul>')

        if request:
            if event.typ != EV_PROTOCOL_REQUEST:
                func = FUNC_MAP[request.typ]
                lines.extend(func(request))
                last = request.data
            request = None
        elif event.typ == EV_REQUEST:
            request = event
            continue
        func = FUNC_MAP[event.typ]
        lines.extend(func(event))
        last = event.typ
    return "\n".join(lines)



"""
    Assertion test module
    ~~~~~~~~~~~~~~~~~~~~~

    :copyright: (c) 2016 by Roland Hedberg.
    :license: APACHE 2.0, see LICENSE for more details.
"""
from future.backports.urllib.parse import parse_qs

import json
import inspect
import traceback
import sys

from otest.events import EV_PROTOCOL_REQUEST
from otest.events import EV_PROTOCOL_RESPONSE
from otest.events import EV_REDIRECT_URL
from otest.events import EV_RESPONSE
from otest.events import EV_HTTP_RESPONSE

from oic.oic import message

__author__ = 'rolandh'

INFORMATION = 0
OK = 1
WARNING = 2
ERROR = 3
CRITICAL = 4
INTERACTION = 5
INCOMPLETE = 6
NOT_APPLICABLE = 7

STATUSCODE = ["INFORMATION", "OK", "WARNING", "ERROR", "CRITICAL",
              "INTERACTION", 'PARTIAL RESULT']

STATUSCODE_TRANSL = dict([(STATUSCODE[i], i) for i in range(len(STATUSCODE))])

END_TAG = "==== END ===="


class TestResult(object):
    name = 'test_result'

    def __init__(self, test_id, status=OK, name='', mti=False, message='',
                 **kwargs):
        self.test_id = test_id
        self.status = status
        self.name = name
        self.mti = mti
        self.message = message
        self.http_status = 0
        self.cid = ''
        self.extra = kwargs

    def __str__(self):
        if self.status:
            return '{}: status={}, message={}'.format(self.test_id,
                                                      STATUSCODE[self.status],
                                                      self.message)
        else:
            return '{}: status=?'.format(self.test_id)


class State(object):
    name = 'state'

    def __init__(self, test_id, status, name='', mti=False, message='',
                 context='', **kwargs):
        self.test_id = test_id
        self.status = status
        self.name = name
        self.mti = mti
        self.message = message
        self.context = context
        self.kwargs = kwargs

    def __str__(self):
        _info = {
            'ctx': self.context, 'id': self.test_id,
            'stat': STATUSCODE[self.status], 'msg': self.message
        }
        if self.status != OK:
            if self.context:
                txt = '{ctx}:{id}: status={stat}, message={msg}'.format(
                    **_info)
            else:
                txt = '{id}: status={stat}, message={msg}'.format(**_info)
        else:
            if self.context:
                txt = '{ctx}:{id}: status={stat}'.format(**_info)
            else:
                txt = '{id}: status={stat}'.format(**_info)

        if self.name:
            txt = '{} [{}]'.format(txt, self.name)

        return txt


class Check(object):
    """ General test
    """

    cid = "check"
    msg = "OK"
    mti = True
    state_cls = State

    def __init__(self, **kwargs):
        self._status = OK
        self._message = ""
        self.content = None
        self.url = ""
        self._kwargs = kwargs

    def _func(self, conv):
        return TestResult('')

    def __call__(self, conv=None, output=None):
        _stat = self._func(conv)
        if isinstance(_stat, dict):
            _stat = self.response(**_stat)

        if output is not None:
            output.append(_stat)
        return _stat

    def response(self, **kwargs):
        try:
            name = " ".join(
                [str(s).strip() for s in self.__doc__.strip().split("\n")])
        except AttributeError:
            name = ""

        res = self.state_cls(test_id=self.cid, status=self._status, name=name,
                             mti=self.mti)

        if self._message:
            res.message = self._message
        else:
            if self._status != OK:
                res.message = self.msg

        for key, val in kwargs.items():
            setattr(self, key, val)

        return res


class ExpectedError(Check):
    pass


class CriticalError(Check):
    status = CRITICAL


class Information(Check):
    status = INFORMATION


class Warnings(Check):
    status = WARNING


class Error(Check):
    status = ERROR


class ResponseInfo(Information):
    """Response information"""

    def _func(self, conv=None):
        self._status = self.status
        _msg = conv.events.last_item(EV_RESPONSE)

        if isinstance(_msg, str):
            self._message = _msg
        else:
            self._message = _msg.to_dict()

        return {}


class WrapException(CriticalError):
    """
    A runtime exception
    """
    cid = "exception"
    msg = "Test tool exception"

    def _func(self, conv=None):
        self._status = self.status
        self._message = traceback.format_exception(*sys.exc_info())
        return {}


class Other(CriticalError):
    """ Other error """
    msg = "Other error"


class CheckHTTPResponse(CriticalError):
    """
    Checks that the HTTP response status is within a specified range
    """
    cid = "http_response"
    msg = "Incorrect HTTP status_code"

    def _func(self, conv):
        _response = conv.events.last_item(EV_HTTP_RESPONSE)

        res = {}
        if not _response:
            return res

        if 'status_code' in self._kwargs:
            if _response.status_code not in self._kwargs['status_code']:
                self._status = self.status
                self._message = self.msg
                res["http_status"] = _response.status_code
        else:
            if _response.status_code >= 400:
                self._status = self.status
                self._message = self.msg
                res["http_status"] = _response.status_code

        return res


def factory(cid):
    for name, obj in inspect.getmembers(sys.modules[__name__]):
        if inspect.isclass(obj):
            try:
                if obj.cid == cid:
                    return obj
            except AttributeError:
                pass

    return None


def get_provider_info(conv):
    _pi = conv.entity.provider_info
    if not _pi:
        _pi = conv.provider_info
    return _pi


def get_protocol_response(conv, cls):
    return conv.events.get_messages(EV_PROTOCOL_RESPONSE, cls)


def get_protocol_request(conv, cls):
    return conv.events.get_messages(EV_PROTOCOL_REQUEST, cls)


def get_id_tokens(conv):
    res = []
    # In access token responses
    for inst in get_protocol_response(conv, message.AccessTokenResponse):
        try:
            res.append(inst["id_token"])
        except KeyError:
            pass

    # implicit, id_token in authorization response
    for inst in get_protocol_response(conv, message.AuthorizationResponse):
        try:
            res.append(inst["id_token"])
        except KeyError:
            pass

    return res


def get_signed_id_tokens(conv):
    res = []
    for item in conv.events.get_data(EV_RESPONSE):
        if isinstance(item, dict):
            ent = item
        else:
            try:
                ent = json.loads(item)
            except Exception as err:
                try:
                    ent = parse_qs(item)
                except:
                    continue
                else:
                    try:
                        res.append(ent['id_token'][0])
                    except KeyError:
                        pass
                    else:
                        continue
        try:
            res.append(ent['id_token'])
        except KeyError:
            pass

    return res


def get_authorization_request(conv, cls):
    authz_req = conv.events.get_data(EV_REDIRECT_URL)[0].split('?')[1]
    return cls().from_urlencoded(authz_req)



from oic import oic
from oic.extension import client
from oic.oic import RegistrationResponse
from oic.utils.authn.client import CLIENT_AUTHN_METHOD
from otest.events import EV_RESPONSE
from otest.events import EV_PROTOCOL_RESPONSE

__author__ = 'roland'


class OicClient(oic.Client):
    def __init__(self, *args, **kwargs):
        oic.Client.__init__(self, *args, **kwargs)
        self.conv = None

    def store_response(self, clinst, text):
        self.conv.events.store(EV_RESPONSE, text)
        self.conv.events.store(EV_PROTOCOL_RESPONSE, clinst)


class ExtClient(client.Client):
    def __init__(self, *args, **kwargs):
        client.Client.__init__(self, *args, **kwargs)
        self.conv = None

    def store_response(self, clinst, text):
        self.conv.events.store(EV_RESPONSE, text)
        self.conv.events.store(EV_PROTOCOL_RESPONSE, clinst)


class Factory(object):
    def __init__(self, client_cls):
        self.client_cls = client_cls

    def adjust_kid(self, kidd, keyjar):
        """
        Verify that the set of keys assigned to different usage is still
        around.
        
        :param kidd: Dictionary {usage: {key_type: kid}} 
        :param keyjar:  A KeyJar instance
        :return: A corrected kid dictionary
        """
        res = {}
        for usage, spec in kidd.items():
            res[usage] = {}
            for key_type, _id in spec.items():
                if not keyjar.get(usage, key_type, kid=_id):
                    l = keyjar.get(usage, key_type)
                    if l:
                        res[usage][key_type] = l[0].kid
                else:
                    res[usage][key_type] = _id
        return res

    def make_client(self, **kw_args):
        """
        Have to get own copy of keyjar

        :param kw_args:
        :return:
        """
        c_keyjar = kw_args["keyjar"].copy()
        args = {'client_authn_method': CLIENT_AUTHN_METHOD, 'keyjar': c_keyjar}
        try:
            args['verify_ssl'] = kw_args['verify_ssl']
        except KeyError:
            pass
        else:
            c_keyjar.verify_ssl = kw_args['verify_ssl']

        _cli = self.client_cls(**args)

        try:
            _kid = kw_args['kid']
        except KeyError:
            pass
        else:
            kw_args['kid'] = self.adjust_kid(_kid, _cli.keyjar)

        c_info = {'keyjar': c_keyjar}
        for arg, val in list(kw_args.items()):
            if arg in ['keyjar']:
                continue
            setattr(_cli, arg, val)
            if arg == 'provider_info':
                _cli.handle_provider_config(val, val['issuer'])
            elif arg == 'registration_response':
                resp = RegistrationResponse(**val)
                _cli.store_registration_info(resp)
            c_info[arg] = val

        return _cli, c_info
